"""
Enterprise Dashboard Generator Module.

Dynamic dashboards, widgets, charts,
and real-time data visualization.

Example:
    # Create dashboard
    dashboard = create_dashboard("System Overview")
    
    # Add widgets
    dashboard.add_widget(
        MetricWidget("CPU Usage", metric="cpu_percent", type=WidgetType.GAUGE)
    )
    dashboard.add_widget(
        ChartWidget("Request Rate", query="requests_total", type=ChartType.LINE)
    )
    dashboard.add_widget(
        TableWidget("Top Endpoints", data=endpoint_stats)
    )
    
    # Render dashboard
    html = dashboard.render_html()
    json_data = dashboard.render_json()
"""

from __future__ import annotations

import json
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class DashboardError(Exception):
    """Dashboard error."""
    pass


class WidgetType(str, Enum):
    """Widget types."""
    METRIC = "metric"
    GAUGE = "gauge"
    COUNTER = "counter"
    SPARKLINE = "sparkline"
    CHART = "chart"
    TABLE = "table"
    TEXT = "text"
    HTML = "html"
    STATUS = "status"
    LIST = "list"
    MAP = "map"
    HEATMAP = "heatmap"


class ChartType(str, Enum):
    """Chart types."""
    LINE = "line"
    BAR = "bar"
    AREA = "area"
    PIE = "pie"
    DONUT = "donut"
    SCATTER = "scatter"
    HISTOGRAM = "histogram"


class RefreshRate(int, Enum):
    """Refresh rates in seconds."""
    REALTIME = 5
    FAST = 15
    NORMAL = 30
    SLOW = 60
    MANUAL = 0


class ColorScheme(str, Enum):
    """Color schemes."""
    DEFAULT = "default"
    DARK = "dark"
    LIGHT = "light"
    COLORFUL = "colorful"
    MONOCHROME = "monochrome"


@dataclass
class DataSource:
    """Data source configuration."""
    name: str
    type: str = "metrics"
    query: str = ""
    refresh_seconds: int = 30
    params: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Threshold:
    """Value threshold."""
    value: float
    color: str = "red"
    label: str = ""


@dataclass
class WidgetPosition:
    """Widget position in grid."""
    row: int = 0
    col: int = 0
    width: int = 4
    height: int = 2


@dataclass
class WidgetStyle:
    """Widget styling."""
    background: str = "#ffffff"
    text_color: str = "#333333"
    border_color: str = "#e0e0e0"
    font_size: str = "14px"
    padding: str = "10px"


# Base widget
@dataclass
class Widget:
    """Base widget."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    title: str = ""
    widget_type: WidgetType = WidgetType.TEXT
    position: WidgetPosition = field(default_factory=WidgetPosition)
    style: WidgetStyle = field(default_factory=WidgetStyle)
    data_source: Optional[DataSource] = None
    thresholds: List[Threshold] = field(default_factory=list)
    visible: bool = True
    description: str = ""
    
    def render(self) -> Dict[str, Any]:
        """Render widget data."""
        return {
            "id": self.id,
            "title": self.title,
            "type": self.widget_type.value,
            "position": {
                "row": self.position.row,
                "col": self.position.col,
                "width": self.position.width,
                "height": self.position.height,
            },
            "style": {
                "background": self.style.background,
                "textColor": self.style.text_color,
                "borderColor": self.style.border_color,
            },
        }


@dataclass
class MetricWidget(Widget):
    """Metric display widget."""
    metric: str = ""
    value: Any = None
    unit: str = ""
    format: str = "{value}"
    trend: Optional[str] = None
    sparkline_data: List[float] = field(default_factory=list)
    
    def __post_init__(self):
        self.widget_type = WidgetType.METRIC
    
    def set_value(self, value: Any) -> None:
        """Set metric value."""
        self.value = value
        if self.sparkline_data:
            self.sparkline_data.append(float(value))
            self.sparkline_data = self.sparkline_data[-50:]  # Keep last 50
    
    def render(self) -> Dict[str, Any]:
        data = super().render()
        data.update({
            "metric": self.metric,
            "value": self.value,
            "unit": self.unit,
            "format": self.format,
            "trend": self.trend,
            "sparkline": self.sparkline_data,
        })
        return data


@dataclass
class GaugeWidget(Widget):
    """Gauge widget."""
    value: float = 0
    min_value: float = 0
    max_value: float = 100
    unit: str = "%"
    
    def __post_init__(self):
        self.widget_type = WidgetType.GAUGE
    
    def set_value(self, value: float) -> None:
        """Set gauge value."""
        self.value = max(self.min_value, min(self.max_value, value))
    
    def get_color(self) -> str:
        """Get color based on thresholds."""
        for threshold in sorted(self.thresholds, key=lambda t: t.value, reverse=True):
            if self.value >= threshold.value:
                return threshold.color
        return "#4caf50"  # Green
    
    def render(self) -> Dict[str, Any]:
        data = super().render()
        data.update({
            "value": self.value,
            "minValue": self.min_value,
            "maxValue": self.max_value,
            "unit": self.unit,
            "color": self.get_color(),
            "percentage": (self.value - self.min_value) / (self.max_value - self.min_value) * 100,
        })
        return data


@dataclass
class ChartWidget(Widget):
    """Chart widget."""
    chart_type: ChartType = ChartType.LINE
    series: List[Dict[str, Any]] = field(default_factory=list)
    labels: List[str] = field(default_factory=list)
    x_axis_label: str = ""
    y_axis_label: str = ""
    show_legend: bool = True
    stacked: bool = False
    
    def __post_init__(self):
        self.widget_type = WidgetType.CHART
    
    def add_series(
        self,
        name: str,
        data: List[float],
        color: Optional[str] = None,
    ) -> None:
        """Add data series."""
        self.series.append({
            "name": name,
            "data": data,
            "color": color,
        })
    
    def set_labels(self, labels: List[str]) -> None:
        """Set axis labels."""
        self.labels = labels
    
    def render(self) -> Dict[str, Any]:
        data = super().render()
        data.update({
            "chartType": self.chart_type.value,
            "series": self.series,
            "labels": self.labels,
            "xAxisLabel": self.x_axis_label,
            "yAxisLabel": self.y_axis_label,
            "showLegend": self.show_legend,
            "stacked": self.stacked,
        })
        return data


@dataclass
class TableWidget(Widget):
    """Table widget."""
    columns: List[Dict[str, str]] = field(default_factory=list)
    rows: List[Dict[str, Any]] = field(default_factory=list)
    sortable: bool = True
    paginated: bool = True
    page_size: int = 10
    
    def __post_init__(self):
        self.widget_type = WidgetType.TABLE
    
    def set_columns(self, columns: List[str]) -> None:
        """Set column definitions."""
        self.columns = [{"key": c, "label": c.title()} for c in columns]
    
    def set_data(self, data: List[Dict[str, Any]]) -> None:
        """Set table data."""
        self.rows = data
    
    def render(self) -> Dict[str, Any]:
        data = super().render()
        data.update({
            "columns": self.columns,
            "rows": self.rows,
            "sortable": self.sortable,
            "paginated": self.paginated,
            "pageSize": self.page_size,
        })
        return data


@dataclass
class StatusWidget(Widget):
    """Status indicator widget."""
    status: str = "unknown"
    message: str = ""
    last_updated: Optional[datetime] = None
    
    STATUS_COLORS = {
        "healthy": "#4caf50",
        "warning": "#ff9800",
        "error": "#f44336",
        "unknown": "#9e9e9e",
    }
    
    def __post_init__(self):
        self.widget_type = WidgetType.STATUS
    
    def set_status(
        self,
        status: str,
        message: str = "",
    ) -> None:
        """Set status."""
        self.status = status
        self.message = message
        self.last_updated = datetime.utcnow()
    
    def render(self) -> Dict[str, Any]:
        data = super().render()
        data.update({
            "status": self.status,
            "message": self.message,
            "color": self.STATUS_COLORS.get(self.status, "#9e9e9e"),
            "lastUpdated": self.last_updated.isoformat() if self.last_updated else None,
        })
        return data


@dataclass
class TextWidget(Widget):
    """Text/markdown widget."""
    content: str = ""
    format: str = "text"  # text, markdown, html
    
    def __post_init__(self):
        self.widget_type = WidgetType.TEXT
    
    def render(self) -> Dict[str, Any]:
        data = super().render()
        data.update({
            "content": self.content,
            "format": self.format,
        })
        return data


@dataclass
class ListWidget(Widget):
    """List widget."""
    items: List[Dict[str, Any]] = field(default_factory=list)
    item_template: str = "{label}: {value}"
    max_items: int = 10
    
    def __post_init__(self):
        self.widget_type = WidgetType.LIST
    
    def add_item(
        self,
        label: str,
        value: Any,
        color: Optional[str] = None,
        icon: Optional[str] = None,
    ) -> None:
        """Add list item."""
        self.items.append({
            "label": label,
            "value": value,
            "color": color,
            "icon": icon,
        })
        self.items = self.items[-self.max_items:]
    
    def render(self) -> Dict[str, Any]:
        data = super().render()
        data.update({
            "items": self.items,
            "itemTemplate": self.item_template,
        })
        return data


# Dashboard
@dataclass
class Dashboard:
    """Dashboard container."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    description: str = ""
    widgets: List[Widget] = field(default_factory=list)
    refresh_rate: RefreshRate = RefreshRate.NORMAL
    color_scheme: ColorScheme = ColorScheme.DEFAULT
    grid_columns: int = 12
    created_at: datetime = field(default_factory=datetime.utcnow)
    updated_at: datetime = field(default_factory=datetime.utcnow)
    tags: List[str] = field(default_factory=list)
    variables: Dict[str, Any] = field(default_factory=dict)
    
    def add_widget(self, widget: Widget) -> None:
        """Add widget to dashboard."""
        self.widgets.append(widget)
        self.updated_at = datetime.utcnow()
    
    def remove_widget(self, widget_id: str) -> bool:
        """Remove widget."""
        for i, widget in enumerate(self.widgets):
            if widget.id == widget_id:
                self.widgets.pop(i)
                self.updated_at = datetime.utcnow()
                return True
        return False
    
    def get_widget(self, widget_id: str) -> Optional[Widget]:
        """Get widget by ID."""
        for widget in self.widgets:
            if widget.id == widget_id:
                return widget
        return None
    
    def render_json(self) -> Dict[str, Any]:
        """Render dashboard as JSON."""
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "widgets": [w.render() for w in self.widgets if w.visible],
            "refreshRate": self.refresh_rate.value,
            "colorScheme": self.color_scheme.value,
            "gridColumns": self.grid_columns,
            "createdAt": self.created_at.isoformat(),
            "updatedAt": self.updated_at.isoformat(),
            "tags": self.tags,
            "variables": self.variables,
        }
    
    def render_html(self) -> str:
        """Render dashboard as HTML."""
        widgets_html = ""
        
        for widget in self.widgets:
            if not widget.visible:
                continue
            
            pos = widget.position
            style = f"""
                grid-column: {pos.col + 1} / span {pos.width};
                grid-row: {pos.row + 1} / span {pos.height};
                background: {widget.style.background};
                color: {widget.style.text_color};
                border: 1px solid {widget.style.border_color};
                padding: {widget.style.padding};
                border-radius: 4px;
            """
            
            content = self._render_widget_content(widget)
            
            widgets_html += f"""
                <div class="widget widget-{widget.widget_type.value}" 
                     id="widget-{widget.id}" style="{style}">
                    <div class="widget-title">{widget.title}</div>
                    <div class="widget-content">{content}</div>
                </div>
            """
        
        return f"""
<!DOCTYPE html>
<html>
<head>
    <title>{self.name}</title>
    <style>
        body {{
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
            margin: 0;
            padding: 20px;
            background: {'#1a1a2e' if self.color_scheme == ColorScheme.DARK else '#f5f5f5'};
            color: {'#ffffff' if self.color_scheme == ColorScheme.DARK else '#333333'};
        }}
        .dashboard-title {{
            font-size: 24px;
            font-weight: bold;
            margin-bottom: 20px;
        }}
        .dashboard-grid {{
            display: grid;
            grid-template-columns: repeat({self.grid_columns}, 1fr);
            gap: 10px;
            auto-rows: minmax(100px, auto);
        }}
        .widget {{
            box-shadow: 0 2px 4px rgba(0,0,0,0.1);
        }}
        .widget-title {{
            font-weight: bold;
            margin-bottom: 10px;
            font-size: 14px;
            opacity: 0.8;
        }}
        .widget-content {{
            font-size: 24px;
        }}
        .gauge-value {{
            font-size: 36px;
            font-weight: bold;
        }}
        .status-indicator {{
            display: inline-block;
            width: 12px;
            height: 12px;
            border-radius: 50%;
            margin-right: 8px;
        }}
        table {{
            width: 100%;
            border-collapse: collapse;
            font-size: 14px;
        }}
        th, td {{
            text-align: left;
            padding: 8px;
            border-bottom: 1px solid #ddd;
        }}
    </style>
</head>
<body>
    <div class="dashboard-title">{self.name}</div>
    <div class="dashboard-grid">
        {widgets_html}
    </div>
    <script>
        // Auto-refresh
        {f'setInterval(() => location.reload(), {self.refresh_rate.value * 1000});' if self.refresh_rate != RefreshRate.MANUAL else ''}
    </script>
</body>
</html>
        """
    
    def _render_widget_content(self, widget: Widget) -> str:
        """Render widget content HTML."""
        if isinstance(widget, MetricWidget):
            return f'<div class="metric-value">{widget.value} {widget.unit}</div>'
        
        elif isinstance(widget, GaugeWidget):
            return f'''
                <div class="gauge-value" style="color: {widget.get_color()}">
                    {widget.value:.1f}{widget.unit}
                </div>
                <div class="gauge-bar" style="
                    width: 100%; 
                    height: 8px; 
                    background: #e0e0e0; 
                    border-radius: 4px;
                ">
                    <div style="
                        width: {(widget.value - widget.min_value) / (widget.max_value - widget.min_value) * 100}%;
                        height: 100%;
                        background: {widget.get_color()};
                        border-radius: 4px;
                    "></div>
                </div>
            '''
        
        elif isinstance(widget, StatusWidget):
            return f'''
                <div>
                    <span class="status-indicator" 
                          style="background: {widget.STATUS_COLORS.get(widget.status, '#9e9e9e')}">
                    </span>
                    {widget.status.upper()}
                </div>
                <div style="font-size: 14px; opacity: 0.8;">{widget.message}</div>
            '''
        
        elif isinstance(widget, TableWidget):
            headers = "".join(f"<th>{c.get('label', c['key'])}</th>" for c in widget.columns)
            rows = ""
            for row in widget.rows[:widget.page_size]:
                cells = "".join(f"<td>{row.get(c['key'], '')}</td>" for c in widget.columns)
                rows += f"<tr>{cells}</tr>"
            return f"<table><thead><tr>{headers}</tr></thead><tbody>{rows}</tbody></table>"
        
        elif isinstance(widget, TextWidget):
            return f'<div style="font-size: 14px;">{widget.content}</div>'
        
        elif isinstance(widget, ListWidget):
            items = "".join(
                f'<li>{item.get("label")}: {item.get("value")}</li>'
                for item in widget.items
            )
            return f'<ul style="margin: 0; padding-left: 20px;">{items}</ul>'
        
        return ""


# Dashboard manager
class DashboardManager:
    """Dashboard management service."""
    
    def __init__(self):
        self._dashboards: Dict[str, Dashboard] = {}
    
    def create(
        self,
        name: str,
        description: str = "",
        **kwargs,
    ) -> Dashboard:
        """Create dashboard."""
        dashboard = Dashboard(
            name=name,
            description=description,
            **kwargs,
        )
        self._dashboards[dashboard.id] = dashboard
        return dashboard
    
    def get(self, dashboard_id: str) -> Optional[Dashboard]:
        """Get dashboard."""
        return self._dashboards.get(dashboard_id)
    
    def list(self) -> List[Dashboard]:
        """List all dashboards."""
        return list(self._dashboards.values())
    
    def delete(self, dashboard_id: str) -> bool:
        """Delete dashboard."""
        if dashboard_id in self._dashboards:
            del self._dashboards[dashboard_id]
            return True
        return False
    
    def clone(self, dashboard_id: str, new_name: str) -> Optional[Dashboard]:
        """Clone dashboard."""
        original = self.get(dashboard_id)
        if not original:
            return None
        
        clone = Dashboard(
            name=new_name,
            description=original.description,
            widgets=[w for w in original.widgets],
            refresh_rate=original.refresh_rate,
            color_scheme=original.color_scheme,
            grid_columns=original.grid_columns,
            tags=original.tags.copy(),
            variables=original.variables.copy(),
        )
        
        self._dashboards[clone.id] = clone
        return clone


# Factory functions
def create_dashboard(
    name: str,
    description: str = "",
    **kwargs,
) -> Dashboard:
    """Create dashboard."""
    return Dashboard(name=name, description=description, **kwargs)


def create_metric_widget(
    title: str,
    metric: str,
    **kwargs,
) -> MetricWidget:
    """Create metric widget."""
    return MetricWidget(title=title, metric=metric, **kwargs)


def create_gauge_widget(
    title: str,
    value: float = 0,
    max_value: float = 100,
    **kwargs,
) -> GaugeWidget:
    """Create gauge widget."""
    return GaugeWidget(title=title, value=value, max_value=max_value, **kwargs)


def create_chart_widget(
    title: str,
    chart_type: ChartType = ChartType.LINE,
    **kwargs,
) -> ChartWidget:
    """Create chart widget."""
    return ChartWidget(title=title, chart_type=chart_type, **kwargs)


def create_table_widget(
    title: str,
    columns: List[str],
    **kwargs,
) -> TableWidget:
    """Create table widget."""
    widget = TableWidget(title=title, **kwargs)
    widget.set_columns(columns)
    return widget


def create_status_widget(
    title: str,
    status: str = "unknown",
    **kwargs,
) -> StatusWidget:
    """Create status widget."""
    widget = StatusWidget(title=title, **kwargs)
    widget.set_status(status)
    return widget


def create_dashboard_manager() -> DashboardManager:
    """Create dashboard manager."""
    return DashboardManager()


__all__ = [
    # Exceptions
    "DashboardError",
    # Enums
    "WidgetType",
    "ChartType",
    "RefreshRate",
    "ColorScheme",
    # Data classes
    "DataSource",
    "Threshold",
    "WidgetPosition",
    "WidgetStyle",
    # Widgets
    "Widget",
    "MetricWidget",
    "GaugeWidget",
    "ChartWidget",
    "TableWidget",
    "StatusWidget",
    "TextWidget",
    "ListWidget",
    # Dashboard
    "Dashboard",
    "DashboardManager",
    # Factory functions
    "create_dashboard",
    "create_metric_widget",
    "create_gauge_widget",
    "create_chart_widget",
    "create_table_widget",
    "create_status_widget",
    "create_dashboard_manager",
]



"""
Enterprise Timeout Module.

Provides request timeouts, deadline propagation, and cancellation tokens
for controlled execution with time limits.

Example:
    # Simple timeout
    @timeout(seconds=30)
    async def slow_operation():
        ...
    
    # Deadline propagation
    deadline = Deadline.from_now(seconds=60)
    async with deadline.scope():
        await operation1()  # Uses remaining time
        await operation2()  # Timeout if deadline passed
    
    # Cancellation tokens
    token = CancellationToken()
    task = asyncio.create_task(long_running(token))
    token.cancel()  # Graceful cancellation
"""

from __future__ import annotations

import asyncio
import time
import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    TypeVar,
    Union,
)
from contextlib import asynccontextmanager, contextmanager
from functools import wraps
from enum import Enum
import logging

logger = logging.getLogger(__name__)

T = TypeVar("T")


class TimeoutError(Exception):
    """Operation timed out."""
    
    def __init__(
        self,
        message: str = "Operation timed out",
        elapsed: Optional[float] = None,
        deadline: Optional[float] = None,
    ):
        super().__init__(message)
        self.elapsed = elapsed
        self.deadline = deadline


class CancellationError(Exception):
    """Operation was cancelled."""
    pass


class DeadlineExceeded(TimeoutError):
    """Deadline has been exceeded."""
    pass


class TimeoutStrategy(str, Enum):
    """Timeout handling strategies."""
    CANCEL = "cancel"  # Cancel the operation
    INTERRUPT = "interrupt"  # Send interrupt signal
    IGNORE = "ignore"  # Log but don't raise


@dataclass
class TimeoutConfig:
    """Configuration for timeout behavior."""
    seconds: float
    strategy: TimeoutStrategy = TimeoutStrategy.CANCEL
    on_timeout: Optional[Callable[[float], None]] = None
    message: str = "Operation timed out"


class CancellationToken:
    """
    Token for cooperative cancellation.
    """
    
    def __init__(self):
        self._cancelled = False
        self._callbacks: List[Callable[[], None]] = []
        self._lock = threading.Lock()
    
    @property
    def is_cancelled(self) -> bool:
        """Check if cancellation was requested."""
        return self._cancelled
    
    def cancel(self) -> None:
        """Request cancellation."""
        with self._lock:
            if self._cancelled:
                return
            
            self._cancelled = True
            
            for callback in self._callbacks:
                try:
                    callback()
                except Exception as e:
                    logger.error(f"Cancellation callback error: {e}")
    
    def register_callback(self, callback: Callable[[], None]) -> None:
        """Register a callback for cancellation."""
        with self._lock:
            if self._cancelled:
                callback()
            else:
                self._callbacks.append(callback)
    
    def throw_if_cancelled(self) -> None:
        """Raise CancellationError if cancelled."""
        if self._cancelled:
            raise CancellationError("Operation was cancelled")
    
    async def wait_for_cancellation(self) -> None:
        """Wait until cancelled."""
        while not self._cancelled:
            await asyncio.sleep(0.1)
    
    @classmethod
    def create_linked(cls, *tokens: 'CancellationToken') -> 'CancellationToken':
        """Create a token that cancels when any parent cancels."""
        linked = cls()
        
        for token in tokens:
            token.register_callback(linked.cancel)
        
        return linked


class CancellationTokenSource:
    """
    Source for creating and managing cancellation tokens.
    """
    
    def __init__(self, timeout: Optional[float] = None):
        """
        Initialize token source.
        
        Args:
            timeout: Optional automatic cancellation timeout
        """
        self._token = CancellationToken()
        self._timeout = timeout
        self._timer: Optional[threading.Timer] = None
        
        if timeout:
            self._start_timer()
    
    @property
    def token(self) -> CancellationToken:
        """Get the cancellation token."""
        return self._token
    
    def _start_timer(self) -> None:
        """Start the timeout timer."""
        if self._timeout:
            self._timer = threading.Timer(self._timeout, self.cancel)
            self._timer.daemon = True
            self._timer.start()
    
    def cancel(self) -> None:
        """Cancel the token."""
        self._token.cancel()
        if self._timer:
            self._timer.cancel()
    
    def dispose(self) -> None:
        """Clean up resources."""
        if self._timer:
            self._timer.cancel()


@dataclass
class Deadline:
    """
    Represents a deadline for operation completion.
    """
    timestamp: float
    created_at: float = field(default_factory=time.time)
    
    @classmethod
    def from_now(cls, seconds: float) -> 'Deadline':
        """Create deadline from now."""
        return cls(timestamp=time.time() + seconds)
    
    @classmethod
    def from_timestamp(cls, timestamp: float) -> 'Deadline':
        """Create deadline from timestamp."""
        return cls(timestamp=timestamp)
    
    @property
    def is_exceeded(self) -> bool:
        """Check if deadline is exceeded."""
        return time.time() > self.timestamp
    
    @property
    def remaining(self) -> float:
        """Get remaining time in seconds."""
        return max(0, self.timestamp - time.time())
    
    @property
    def elapsed(self) -> float:
        """Get elapsed time since creation."""
        return time.time() - self.created_at
    
    def extend(self, seconds: float) -> 'Deadline':
        """Create a new deadline extended by seconds."""
        return Deadline(
            timestamp=self.timestamp + seconds,
            created_at=self.created_at,
        )
    
    def check(self) -> None:
        """Raise if deadline exceeded."""
        if self.is_exceeded:
            raise DeadlineExceeded(
                f"Deadline exceeded by {-self.remaining:.2f}s",
                elapsed=self.elapsed,
                deadline=self.timestamp,
            )
    
    @asynccontextmanager
    async def scope(self):
        """
        Context manager that enforces the deadline.
        
        Example:
            async with deadline.scope():
                await operation()
        """
        self.check()
        
        try:
            yield self
        except asyncio.TimeoutError:
            raise DeadlineExceeded(
                f"Deadline exceeded",
                elapsed=self.elapsed,
                deadline=self.timestamp,
            )
    
    async def run(self, coro: Any) -> Any:
        """
        Run a coroutine with this deadline.
        
        Args:
            coro: Coroutine to run
            
        Returns:
            Result of coroutine
        """
        self.check()
        
        try:
            return await asyncio.wait_for(coro, timeout=self.remaining)
        except asyncio.TimeoutError:
            raise DeadlineExceeded(
                f"Deadline exceeded",
                elapsed=self.elapsed,
                deadline=self.timestamp,
            )


# Context variable for deadline propagation
_deadline_context: Dict[int, Deadline] = {}


def get_current_deadline() -> Optional[Deadline]:
    """Get the current context deadline."""
    task = asyncio.current_task()
    if task:
        return _deadline_context.get(id(task))
    return None


def set_current_deadline(deadline: Deadline) -> None:
    """Set the current context deadline."""
    task = asyncio.current_task()
    if task:
        _deadline_context[id(task)] = deadline


def clear_current_deadline() -> None:
    """Clear the current context deadline."""
    task = asyncio.current_task()
    if task:
        _deadline_context.pop(id(task), None)


@asynccontextmanager
async def deadline_scope(seconds: float):
    """
    Create a deadline scope.
    
    Example:
        async with deadline_scope(30):
            await operation()
    """
    deadline = Deadline.from_now(seconds)
    
    # Propagate to nested calls
    old = get_current_deadline()
    
    # Use the shorter deadline
    if old and old.remaining < deadline.remaining:
        deadline = old
    
    set_current_deadline(deadline)
    
    try:
        async with deadline.scope():
            yield deadline
    finally:
        if old:
            set_current_deadline(old)
        else:
            clear_current_deadline()


def timeout(
    seconds: float,
    strategy: TimeoutStrategy = TimeoutStrategy.CANCEL,
    on_timeout: Optional[Callable[[float], None]] = None,
) -> Callable:
    """
    Decorator for function timeout.
    
    Example:
        @timeout(30)
        async def slow_operation():
            ...
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            start = time.time()
            
            try:
                return await asyncio.wait_for(
                    func(*args, **kwargs),
                    timeout=seconds,
                )
            except asyncio.TimeoutError:
                elapsed = time.time() - start
                
                if on_timeout:
                    on_timeout(elapsed)
                
                if strategy == TimeoutStrategy.IGNORE:
                    logger.warning(
                        f"{func.__name__} timed out after {elapsed:.2f}s"
                    )
                    return None
                
                raise TimeoutError(
                    f"{func.__name__} timed out after {elapsed:.2f}s",
                    elapsed=elapsed,
                    deadline=start + seconds,
                )
        
        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            import concurrent.futures
            
            start = time.time()
            
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(func, *args, **kwargs)
                try:
                    return future.result(timeout=seconds)
                except concurrent.futures.TimeoutError:
                    elapsed = time.time() - start
                    
                    if on_timeout:
                        on_timeout(elapsed)
                    
                    if strategy == TimeoutStrategy.IGNORE:
                        logger.warning(
                            f"{func.__name__} timed out after {elapsed:.2f}s"
                        )
                        return None
                    
                    raise TimeoutError(
                        f"{func.__name__} timed out after {elapsed:.2f}s",
                        elapsed=elapsed,
                    )
        
        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
    
    return decorator


def with_deadline(seconds: float) -> Callable:
    """
    Decorator that enforces a deadline.
    
    Example:
        @with_deadline(60)
        async def operation():
            ...
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            async with deadline_scope(seconds):
                return await func(*args, **kwargs)
        
        return wrapper
    
    return decorator


def with_cancellation(token_arg: str = "token") -> Callable:
    """
    Decorator that supports cancellation.
    
    Example:
        @with_cancellation()
        async def operation(token: CancellationToken):
            while not token.is_cancelled:
                await do_work()
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            token = kwargs.get(token_arg)
            
            if token and isinstance(token, CancellationToken):
                # Check before starting
                token.throw_if_cancelled()
            
            return await func(*args, **kwargs)
        
        return wrapper
    
    return decorator


class TimeoutBudget:
    """
    Budget for distributing timeout across operations.
    """
    
    def __init__(self, total_seconds: float):
        """
        Initialize timeout budget.
        
        Args:
            total_seconds: Total time budget
        """
        self.total_seconds = total_seconds
        self._start_time = time.time()
        self._used: Dict[str, float] = {}
    
    @property
    def remaining(self) -> float:
        """Get remaining budget."""
        elapsed = time.time() - self._start_time
        return max(0, self.total_seconds - elapsed)
    
    @property
    def is_exhausted(self) -> bool:
        """Check if budget is exhausted."""
        return self.remaining <= 0
    
    def allocate(self, operation: str, fraction: float = 0.5) -> float:
        """
        Allocate a portion of remaining budget.
        
        Args:
            operation: Operation name for tracking
            fraction: Fraction of remaining to allocate (0-1)
            
        Returns:
            Allocated timeout in seconds
        """
        if self.is_exhausted:
            raise TimeoutError("Timeout budget exhausted")
        
        allocated = self.remaining * fraction
        self._used[operation] = allocated
        return allocated
    
    def check(self) -> None:
        """Raise if budget exhausted."""
        if self.is_exhausted:
            raise TimeoutError(
                f"Timeout budget exhausted after {self.total_seconds}s",
                elapsed=time.time() - self._start_time,
            )
    
    @asynccontextmanager
    async def operation(self, name: str, fraction: float = 0.5):
        """
        Context manager for a budgeted operation.
        
        Example:
            async with budget.operation("api_call", 0.3):
                await api.call()
        """
        timeout_seconds = self.allocate(name, fraction)
        start = time.time()
        
        try:
            yield timeout_seconds
        finally:
            elapsed = time.time() - start
            self._used[name] = elapsed


class AdaptiveTimeout:
    """
    Adaptive timeout based on historical data.
    """
    
    def __init__(
        self,
        initial: float = 30.0,
        min_timeout: float = 5.0,
        max_timeout: float = 120.0,
        percentile: float = 95.0,
    ):
        """
        Initialize adaptive timeout.
        
        Args:
            initial: Initial timeout
            min_timeout: Minimum timeout
            max_timeout: Maximum timeout
            percentile: Target percentile for timeout calculation
        """
        self.initial = initial
        self.min_timeout = min_timeout
        self.max_timeout = max_timeout
        self.percentile = percentile
        
        self._latencies: List[float] = []
        self._current = initial
    
    @property
    def current(self) -> float:
        """Get current timeout value."""
        return self._current
    
    def record(self, latency: float) -> None:
        """Record a latency observation."""
        self._latencies.append(latency)
        
        # Keep last 100 observations
        if len(self._latencies) > 100:
            self._latencies.pop(0)
        
        self._recalculate()
    
    def _recalculate(self) -> None:
        """Recalculate timeout based on observations."""
        if len(self._latencies) < 10:
            return
        
        sorted_latencies = sorted(self._latencies)
        index = int(len(sorted_latencies) * self.percentile / 100)
        percentile_latency = sorted_latencies[min(index, len(sorted_latencies) - 1)]
        
        # Add margin
        new_timeout = percentile_latency * 1.5
        
        self._current = max(
            self.min_timeout,
            min(self.max_timeout, new_timeout),
        )
    
    def __call__(self) -> float:
        """Get current timeout."""
        return self._current


async def race(*coros: Any, return_when: str = "FIRST_COMPLETED") -> Any:
    """
    Race multiple coroutines, return first to complete.
    
    Example:
        result = await race(
            fetch_from_primary(),
            fetch_from_secondary(),
        )
    """
    tasks = [asyncio.create_task(coro) for coro in coros]
    
    try:
        done, pending = await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_COMPLETED,
        )
        
        # Cancel remaining
        for task in pending:
            task.cancel()
        
        # Return first result
        for task in done:
            return task.result()
            
    except Exception:
        # Cancel all on error
        for task in tasks:
            task.cancel()
        raise


__all__ = [
    # Exceptions
    "TimeoutError",
    "CancellationError",
    "DeadlineExceeded",
    # Enums
    "TimeoutStrategy",
    # Data classes
    "TimeoutConfig",
    "Deadline",
    # Cancellation
    "CancellationToken",
    "CancellationTokenSource",
    # Budget
    "TimeoutBudget",
    "AdaptiveTimeout",
    # Decorators
    "timeout",
    "with_deadline",
    "with_cancellation",
    # Context managers
    "deadline_scope",
    # Utility functions
    "get_current_deadline",
    "set_current_deadline",
    "clear_current_deadline",
    "race",
]



"""
Remote Agent Communication - Client and Server.

Enables agents to:
- Expose themselves as remote services
- Call other remote agents
- Handle multiple protocols (HTTP, SSE, MQTT, WebSocket)
"""

import json
import uuid
import time
import logging
import threading
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterator, List, Optional, Type, Union
from enum import Enum
from datetime import datetime
from queue import Queue

from .protocols import (
    ProtocolType,
    ProtocolConfig,
    CommunicationProtocol,
    HTTPProtocol,
    SSEProtocol,
    MQTTProtocol,
    WebSocketProtocol,
    STDIOProtocol,
)

logger = logging.getLogger(__name__)


class MessageType(Enum):
    """Types of agent messages."""
    QUERY = "query"
    RESPONSE = "response"
    STREAM = "stream"
    STREAM_END = "stream_end"
    HANDOFF = "handoff"
    HEARTBEAT = "heartbeat"
    ERROR = "error"
    ACK = "ack"


@dataclass
class AgentMessage:
    """Message structure for agent communication."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    type: MessageType = MessageType.QUERY
    sender: str = ""
    recipient: str = ""
    content: Any = None
    context: Dict[str, Any] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)
    timestamp: str = field(default_factory=lambda: datetime.now().isoformat())
    reply_to: Optional[str] = None
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "id": self.id,
            "type": self.type.value,
            "sender": self.sender,
            "recipient": self.recipient,
            "content": self.content,
            "context": self.context,
            "metadata": self.metadata,
            "timestamp": self.timestamp,
            "reply_to": self.reply_to,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "AgentMessage":
        """Create from dictionary."""
        return cls(
            id=data.get("id", str(uuid.uuid4())),
            type=MessageType(data.get("type", "query")),
            sender=data.get("sender", ""),
            recipient=data.get("recipient", ""),
            content=data.get("content"),
            context=data.get("context", {}),
            metadata=data.get("metadata", {}),
            timestamp=data.get("timestamp", datetime.now().isoformat()),
            reply_to=data.get("reply_to"),
        )


@dataclass
class AgentEndpoint:
    """Configuration for a remote agent endpoint."""
    agent_id: str
    protocol: ProtocolType = ProtocolType.HTTP
    host: str = "localhost"
    port: int = 8080
    path: str = "/agent"
    auth_token: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    # MQTT-specific
    mqtt_topic: Optional[str] = None
    
    # STDIO-specific
    command: Optional[List[str]] = None


class RemoteAgentClient:
    """
    Client for calling remote agents over various protocols.
    
    Example:
        >>> # HTTP/HTTPS
        >>> client = RemoteAgentClient()
        >>> client.register_endpoint(AgentEndpoint(
        ...     agent_id="analyzer",
        ...     protocol=ProtocolType.HTTPS,
        ...     host="analyzer.example.com",
        ...     port=443
        ... ))
        >>> response = client.call("analyzer", "Analyze this document")
        
        >>> # MQTT
        >>> client.register_endpoint(AgentEndpoint(
        ...     agent_id="iot-agent",
        ...     protocol=ProtocolType.MQTT,
        ...     host="mqtt.example.com",
        ...     mqtt_topic="agents/iot"
        ... ))
        >>> response = client.call("iot-agent", "Get sensor data")
        
        >>> # Streaming with SSE
        >>> for chunk in client.call_stream("summarizer", "Summarize article"):
        ...     print(chunk)
    """
    
    def __init__(self, agent_id: Optional[str] = None):
        self.agent_id = agent_id or f"client-{uuid.uuid4().hex[:8]}"
        self._endpoints: Dict[str, AgentEndpoint] = {}
        self._protocols: Dict[str, CommunicationProtocol] = {}
        self._connected: Dict[str, bool] = {}
    
    def register_endpoint(self, endpoint: AgentEndpoint) -> None:
        """Register a remote agent endpoint."""
        self._endpoints[endpoint.agent_id] = endpoint
        logger.info(f"Registered endpoint: {endpoint.agent_id} ({endpoint.protocol.value})")
    
    def register_endpoints(self, endpoints: List[AgentEndpoint]) -> None:
        """Register multiple endpoints."""
        for endpoint in endpoints:
            self.register_endpoint(endpoint)
    
    def connect(self, agent_id: str) -> bool:
        """Connect to a specific remote agent."""
        if agent_id not in self._endpoints:
            logger.error(f"Unknown agent: {agent_id}")
            return False
        
        endpoint = self._endpoints[agent_id]
        protocol = self._create_protocol(endpoint)
        
        if protocol.connect():
            self._protocols[agent_id] = protocol
            self._connected[agent_id] = True
            logger.info(f"Connected to: {agent_id}")
            return True
        return False
    
    def disconnect(self, agent_id: str) -> bool:
        """Disconnect from a remote agent."""
        if agent_id in self._protocols:
            result = self._protocols[agent_id].disconnect()
            del self._protocols[agent_id]
            self._connected[agent_id] = False
            return result
        return True
    
    def disconnect_all(self) -> None:
        """Disconnect from all agents."""
        for agent_id in list(self._protocols.keys()):
            self.disconnect(agent_id)
    
    def call(
        self,
        agent_id: str,
        prompt: Union[str, Dict[str, Any]],
        context: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
    ) -> Dict[str, Any]:
        """
        Call a remote agent and get response.
        
        Args:
            agent_id: Target agent identifier
            prompt: Query or message to send
            context: Additional context
            timeout: Response timeout
            
        Returns:
            Agent response dictionary
        """
        # Auto-connect if needed
        if agent_id not in self._protocols:
            if not self.connect(agent_id):
                return {"error": f"Cannot connect to agent: {agent_id}"}
        
        protocol = self._protocols[agent_id]
        
        # Build message
        message = AgentMessage(
            type=MessageType.QUERY,
            sender=self.agent_id,
            recipient=agent_id,
            content=prompt if isinstance(prompt, dict) else {"prompt": prompt},
            context=context or {},
        )
        
        # Set timeout if specified
        if timeout:
            protocol.config.timeout = timeout
        
        # Send and get response
        response = protocol.send(message.to_dict())
        
        return response
    
    def call_stream(
        self,
        agent_id: str,
        prompt: Union[str, Dict[str, Any]],
        context: Optional[Dict[str, Any]] = None,
    ) -> Iterator[Dict[str, Any]]:
        """
        Call a remote agent and stream the response.
        
        Uses SSE or WebSocket for streaming.
        """
        if agent_id not in self._endpoints:
            yield {"error": f"Unknown agent: {agent_id}"}
            return
        
        endpoint = self._endpoints[agent_id]
        
        # Use SSE for streaming
        if endpoint.protocol in [ProtocolType.SSE, ProtocolType.HTTP, ProtocolType.HTTPS]:
            protocol = SSEProtocol(
                host=endpoint.host,
                port=endpoint.port,
                path=endpoint.path,
                use_ssl=(endpoint.protocol == ProtocolType.HTTPS),
            )
            
            if not protocol.connect():
                yield {"error": "Failed to connect for streaming"}
                return
            
            message = AgentMessage(
                type=MessageType.QUERY,
                sender=self.agent_id,
                recipient=agent_id,
                content=prompt if isinstance(prompt, dict) else {"prompt": prompt},
                context=context or {},
                metadata={"stream": True},
            )
            
            try:
                for event in protocol.send_and_stream(message.to_dict()):
                    yield event
            finally:
                protocol.disconnect()
        
        # Use WebSocket for streaming
        elif endpoint.protocol == ProtocolType.WEBSOCKET:
            if agent_id not in self._protocols:
                if not self.connect(agent_id):
                    yield {"error": "Failed to connect"}
                    return
            
            protocol = self._protocols[agent_id]
            
            message = AgentMessage(
                type=MessageType.QUERY,
                sender=self.agent_id,
                recipient=agent_id,
                content=prompt if isinstance(prompt, dict) else {"prompt": prompt},
                context=context or {},
                metadata={"stream": True},
            )
            
            protocol.send_nowait(message.to_dict())
            
            while True:
                response = protocol.receive(timeout=30.0)
                if not response:
                    break
                yield response
                if response.get("type") == MessageType.STREAM_END.value:
                    break
    
    def handoff(
        self,
        from_agent: str,
        to_agent: str,
        context: Dict[str, Any],
        reason: str = "",
    ) -> Dict[str, Any]:
        """
        Handoff execution from one agent to another.
        
        Args:
            from_agent: Source agent
            to_agent: Target agent
            context: Conversation/execution context to transfer
            reason: Reason for handoff
        """
        message = AgentMessage(
            type=MessageType.HANDOFF,
            sender=from_agent,
            recipient=to_agent,
            content={
                "reason": reason,
                "handoff_context": context,
            },
            context=context,
        )
        
        return self.call(to_agent, message.to_dict())
    
    def broadcast(
        self,
        agent_ids: List[str],
        prompt: Union[str, Dict[str, Any]],
        context: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Dict[str, Any]]:
        """
        Broadcast message to multiple agents.
        
        Returns:
            Dictionary mapping agent_id to response
        """
        results = {}
        for agent_id in agent_ids:
            results[agent_id] = self.call(agent_id, prompt, context)
        return results
    
    def _create_protocol(self, endpoint: AgentEndpoint) -> CommunicationProtocol:
        """Create protocol instance for endpoint."""
        config = ProtocolConfig(
            protocol_type=endpoint.protocol,
            host=endpoint.host,
            port=endpoint.port,
            path=endpoint.path,
            auth_token=endpoint.auth_token,
        )
        
        if endpoint.protocol == ProtocolType.HTTP:
            return HTTPProtocol(
                host=endpoint.host,
                port=endpoint.port,
                path=endpoint.path,
                use_ssl=False,
                config=config,
            )
        
        elif endpoint.protocol == ProtocolType.HTTPS:
            return HTTPProtocol(
                host=endpoint.host,
                port=endpoint.port,
                path=endpoint.path,
                use_ssl=True,
                config=config,
            )
        
        elif endpoint.protocol == ProtocolType.SSE:
            return SSEProtocol(
                host=endpoint.host,
                port=endpoint.port,
                path=endpoint.path,
                config=config,
            )
        
        elif endpoint.protocol == ProtocolType.MQTT:
            return MQTTProtocol(
                broker=endpoint.host,
                port=endpoint.port,
                topic=endpoint.mqtt_topic or f"agents/{endpoint.agent_id}",
                config=config,
            )
        
        elif endpoint.protocol == ProtocolType.WEBSOCKET:
            return WebSocketProtocol(
                host=endpoint.host,
                port=endpoint.port,
                path=endpoint.path,
                config=config,
            )
        
        elif endpoint.protocol == ProtocolType.STDIO:
            return STDIOProtocol(
                command=endpoint.command,
                config=config,
            )
        
        else:
            raise ValueError(f"Unsupported protocol: {endpoint.protocol}")


class RemoteAgentServer:
    """
    Server to expose an agent as a remote service.
    
    Example:
        >>> from agenticaiframework import Agent
        >>> 
        >>> agent = Agent(name="my-agent", model="gpt-4")
        >>> 
        >>> server = RemoteAgentServer(agent)
        >>> server.start(host="0.0.0.0", port=8080)
        
        >>> # With Flask
        >>> app = server.create_flask_app()
        >>> app.run(port=8080)
        
        >>> # With FastAPI
        >>> app = server.create_fastapi_app()
        >>> uvicorn.run(app, port=8080)
    """
    
    def __init__(
        self,
        agent: Any,  # Agent instance
        agent_id: Optional[str] = None,
    ):
        self.agent = agent
        self.agent_id = agent_id or getattr(agent, "name", str(uuid.uuid4()))
        self._handlers: Dict[MessageType, Callable] = {}
        self._running = False
        
        # Register default handlers
        self._setup_default_handlers()
    
    def _setup_default_handlers(self) -> None:
        """Set up default message handlers."""
        self._handlers[MessageType.QUERY] = self._handle_query
        self._handlers[MessageType.HANDOFF] = self._handle_handoff
        self._handlers[MessageType.HEARTBEAT] = self._handle_heartbeat
    
    def on_message(
        self,
        message_type: MessageType,
    ) -> Callable:
        """Decorator to register custom message handlers."""
        def decorator(func: Callable) -> Callable:
            self._handlers[message_type] = func
            return func
        return decorator
    
    def handle_message(self, message: Dict[str, Any]) -> Dict[str, Any]:
        """Process incoming message and return response."""
        try:
            agent_msg = AgentMessage.from_dict(message)
            handler = self._handlers.get(agent_msg.type, self._handle_query)
            return handler(agent_msg)
        except Exception as e:
            logger.error(f"Message handling error: {e}")
            return {
                "type": MessageType.ERROR.value,
                "error": str(e),
            }
    
    def _handle_query(self, message: AgentMessage) -> Dict[str, Any]:
        """Handle query messages."""
        try:
            content = message.content
            prompt = content.get("prompt") if isinstance(content, dict) else str(content)
            
            # Get context
            context = message.context
            
            # Call agent
            if hasattr(self.agent, "run"):
                result = self.agent.run(prompt, context=context)
            elif hasattr(self.agent, "invoke"):
                result = self.agent.invoke(prompt)
            elif callable(self.agent):
                result = self.agent(prompt)
            else:
                result = {"error": "Agent not callable"}
            
            return AgentMessage(
                type=MessageType.RESPONSE,
                sender=self.agent_id,
                recipient=message.sender,
                content=result,
                reply_to=message.id,
            ).to_dict()
            
        except Exception as e:
            return {
                "type": MessageType.ERROR.value,
                "error": str(e),
                "reply_to": message.id,
            }
    
    def _handle_handoff(self, message: AgentMessage) -> Dict[str, Any]:
        """Handle handoff messages."""
        try:
            content = message.content
            handoff_context = content.get("handoff_context", {})
            
            # Update agent's context if supported
            if hasattr(self.agent, "set_context"):
                self.agent.set_context(handoff_context)
            
            return AgentMessage(
                type=MessageType.ACK,
                sender=self.agent_id,
                recipient=message.sender,
                content={"status": "handoff_accepted"},
                reply_to=message.id,
            ).to_dict()
            
        except Exception as e:
            return {"type": MessageType.ERROR.value, "error": str(e)}
    
    def _handle_heartbeat(self, message: AgentMessage) -> Dict[str, Any]:
        """Handle heartbeat messages."""
        return AgentMessage(
            type=MessageType.ACK,
            sender=self.agent_id,
            recipient=message.sender,
            content={"status": "alive", "timestamp": datetime.now().isoformat()},
            reply_to=message.id,
        ).to_dict()
    
    def create_flask_app(self):
        """Create Flask app for HTTP server."""
        try:
            from flask import Flask, request, jsonify, Response
        except ImportError:
            raise ImportError("Flask required: pip install flask")
        
        app = Flask(f"agent-{self.agent_id}")
        
        @app.route("/health", methods=["GET"])
        def health():
            return jsonify({"status": "healthy", "agent_id": self.agent_id})
        
        @app.route("/agent", methods=["POST"])
        def agent_endpoint():
            message = request.json
            response = self.handle_message(message)
            return jsonify(response)
        
        @app.route("/agent/stream", methods=["POST"])
        def agent_stream():
            message = request.json
            
            def generate():
                agent_msg = AgentMessage.from_dict(message)
                content = agent_msg.content
                prompt = content.get("prompt") if isinstance(content, dict) else str(content)
                
                if hasattr(self.agent, "stream"):
                    for chunk in self.agent.stream(prompt):
                        yield f"data: {json.dumps({'chunk': chunk})}\n\n"
                else:
                    result = self._handle_query(agent_msg)
                    yield f"data: {json.dumps(result)}\n\n"
                
                yield f"data: {json.dumps({'type': 'stream_end'})}\n\n"
            
            return Response(generate(), mimetype="text/event-stream")
        
        return app
    
    def create_fastapi_app(self):
        """Create FastAPI app for HTTP server."""
        try:
            from fastapi import FastAPI
            from fastapi.responses import StreamingResponse
            from pydantic import BaseModel
        except ImportError:
            raise ImportError("FastAPI required: pip install fastapi uvicorn")
        
        app = FastAPI(title=f"Agent: {self.agent_id}")
        
        @app.get("/health")
        def health():
            return {"status": "healthy", "agent_id": self.agent_id}
        
        @app.post("/agent")
        def agent_endpoint(message: dict):
            return self.handle_message(message)
        
        @app.post("/agent/stream")
        def agent_stream(message: dict):
            async def generate():
                agent_msg = AgentMessage.from_dict(message)
                content = agent_msg.content
                prompt = content.get("prompt") if isinstance(content, dict) else str(content)
                
                if hasattr(self.agent, "stream"):
                    for chunk in self.agent.stream(prompt):
                        yield f"data: {json.dumps({'chunk': chunk})}\n\n"
                else:
                    result = self._handle_query(agent_msg)
                    yield f"data: {json.dumps(result)}\n\n"
                
                yield f"data: {json.dumps({'type': 'stream_end'})}\n\n"
            
            return StreamingResponse(generate(), media_type="text/event-stream")
        
        return app
    
    def create_mqtt_handler(
        self,
        broker: str = "localhost",
        port: int = 1883,
        topic: Optional[str] = None,
    ):
        """Create MQTT message handler."""
        protocol = MQTTProtocol(
            broker=broker,
            port=port,
            topic=topic or f"agents/{self.agent_id}/requests",
        )
        
        response_topic = f"agents/{self.agent_id}/responses"
        
        def on_message(message: Dict[str, Any]) -> None:
            payload = message.get("payload", {})
            response = self.handle_message(payload)
            protocol.send(response, topic=response_topic)
        
        protocol.on_message(on_message)
        protocol.connect()
        protocol.subscribe(topic or f"agents/{self.agent_id}/requests")
        
        return protocol
    
    def start_stdio(self) -> None:
        """Start STDIO server for subprocess communication."""
        import sys
        
        self._running = True
        logger.info(f"STDIO server started for agent: {self.agent_id}")
        
        while self._running:
            try:
                line = sys.stdin.readline()
                if not line:
                    break
                
                message = json.loads(line.strip())
                response = self.handle_message(message)
                
                sys.stdout.write(json.dumps(response) + "\n")
                sys.stdout.flush()
                
            except json.JSONDecodeError:
                pass
            except KeyboardInterrupt:
                break
            except Exception as e:
                error_response = {"type": "error", "error": str(e)}
                sys.stdout.write(json.dumps(error_response) + "\n")
                sys.stdout.flush()
    
    def stop(self) -> None:
        """Stop the server."""
        self._running = False


__all__ = [
    "MessageType",
    "AgentMessage",
    "AgentEndpoint",
    "RemoteAgentClient",
    "RemoteAgentServer",
]



"""
Enterprise Leader Election Module.

Provides leader election, distributed locks, and coordination
primitives for distributed systems.

Example:
    # Create leader elector
    elector = create_leader_elector("my-service")
    
    # Participate in election
    async with elector.campaign() as leadership:
        if leadership.is_leader:
            # I'm the leader
            await do_leader_work()
    
    # Distributed lock
    lock = create_distributed_lock("resource_123")
    async with lock.acquire():
        # Exclusive access to resource
        ...
    
    # With decorator
    @leader_only(elector)
    async def leader_task():
        ...
"""

from __future__ import annotations

import asyncio
import logging
import os
import time
import uuid
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    TypeVar,
)

logger = logging.getLogger(__name__)

T = TypeVar('T')


class LeaderError(Exception):
    """Leader election error."""
    pass


class LockError(Exception):
    """Distributed lock error."""
    pass


class LockAcquisitionError(LockError):
    """Failed to acquire lock."""
    pass


class LeadershipState(str, Enum):
    """Leadership state."""
    FOLLOWER = "follower"
    CANDIDATE = "candidate"
    LEADER = "leader"


class LockState(str, Enum):
    """Lock state."""
    UNLOCKED = "unlocked"
    LOCKED = "locked"
    WAITING = "waiting"


@dataclass
class LeaderInfo:
    """Information about the current leader."""
    leader_id: str
    node_id: str
    elected_at: datetime
    term: int
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Leadership:
    """Leadership context."""
    is_leader: bool
    leader_id: Optional[str] = None
    term: int = 0
    since: Optional[datetime] = None
    
    def assert_leader(self) -> None:
        """Assert that we are the leader."""
        if not self.is_leader:
            raise LeaderError("Not the leader")


@dataclass
class LockInfo:
    """Distributed lock information."""
    lock_id: str
    resource_id: str
    holder_id: str
    acquired_at: datetime
    ttl_seconds: int
    expires_at: datetime


@dataclass
class ElectionConfig:
    """Leader election configuration."""
    election_timeout_ms: int = 5000
    heartbeat_interval_ms: int = 1000
    lease_duration_seconds: int = 30


class LeaderStore(ABC):
    """Abstract leader state store."""
    
    @abstractmethod
    async def try_acquire_leadership(
        self,
        election_name: str,
        node_id: str,
        ttl_seconds: int,
    ) -> bool:
        """Try to acquire leadership."""
        pass
    
    @abstractmethod
    async def renew_leadership(
        self,
        election_name: str,
        node_id: str,
        ttl_seconds: int,
    ) -> bool:
        """Renew leadership lease."""
        pass
    
    @abstractmethod
    async def release_leadership(
        self,
        election_name: str,
        node_id: str,
    ) -> bool:
        """Release leadership."""
        pass
    
    @abstractmethod
    async def get_leader(
        self,
        election_name: str,
    ) -> Optional[LeaderInfo]:
        """Get current leader."""
        pass


class InMemoryLeaderStore(LeaderStore):
    """In-memory leader store (single-node only)."""
    
    def __init__(self):
        self._leaders: Dict[str, LeaderInfo] = {}
        self._lock = asyncio.Lock()
        self._terms: Dict[str, int] = {}
    
    async def try_acquire_leadership(
        self,
        election_name: str,
        node_id: str,
        ttl_seconds: int,
    ) -> bool:
        async with self._lock:
            current = self._leaders.get(election_name)
            
            # Check if current leader's lease expired
            if current:
                if current.elected_at + timedelta(seconds=ttl_seconds) > datetime.now():
                    return current.node_id == node_id
            
            # Acquire leadership
            term = self._terms.get(election_name, 0) + 1
            self._terms[election_name] = term
            
            self._leaders[election_name] = LeaderInfo(
                leader_id=f"{election_name}-leader",
                node_id=node_id,
                elected_at=datetime.now(),
                term=term,
            )
            
            return True
    
    async def renew_leadership(
        self,
        election_name: str,
        node_id: str,
        ttl_seconds: int,
    ) -> bool:
        async with self._lock:
            current = self._leaders.get(election_name)
            
            if current and current.node_id == node_id:
                current.elected_at = datetime.now()
                return True
            
            return False
    
    async def release_leadership(
        self,
        election_name: str,
        node_id: str,
    ) -> bool:
        async with self._lock:
            current = self._leaders.get(election_name)
            
            if current and current.node_id == node_id:
                del self._leaders[election_name]
                return True
            
            return False
    
    async def get_leader(
        self,
        election_name: str,
    ) -> Optional[LeaderInfo]:
        return self._leaders.get(election_name)


class LeaderElector:
    """
    Leader election coordinator.
    """
    
    def __init__(
        self,
        election_name: str,
        store: LeaderStore,
        node_id: Optional[str] = None,
        config: Optional[ElectionConfig] = None,
    ):
        self._election_name = election_name
        self._store = store
        self._node_id = node_id or f"{os.getpid()}-{uuid.uuid4().hex[:8]}"
        self._config = config or ElectionConfig()
        self._state = LeadershipState.FOLLOWER
        self._leadership: Optional[Leadership] = None
        self._callbacks: List[Callable[[Leadership], None]] = []
        self._running = False
        self._heartbeat_task: Optional[asyncio.Task] = None
    
    @property
    def node_id(self) -> str:
        return self._node_id
    
    @property
    def state(self) -> LeadershipState:
        return self._state
    
    @property
    def is_leader(self) -> bool:
        return self._state == LeadershipState.LEADER
    
    async def start(self) -> None:
        """Start participating in election."""
        if self._running:
            return
        
        self._running = True
        self._heartbeat_task = asyncio.create_task(self._election_loop())
        
        logger.info(f"Node {self._node_id} started leader election")
    
    async def stop(self) -> None:
        """Stop participating in election."""
        self._running = False
        
        if self._heartbeat_task:
            self._heartbeat_task.cancel()
            try:
                await self._heartbeat_task
            except asyncio.CancelledError:
                pass
        
        if self._state == LeadershipState.LEADER:
            await self._store.release_leadership(
                self._election_name,
                self._node_id,
            )
        
        self._state = LeadershipState.FOLLOWER
        
        logger.info(f"Node {self._node_id} stopped leader election")
    
    async def _election_loop(self) -> None:
        """Main election loop."""
        while self._running:
            try:
                if self._state == LeadershipState.LEADER:
                    # Renew lease
                    renewed = await self._store.renew_leadership(
                        self._election_name,
                        self._node_id,
                        self._config.lease_duration_seconds,
                    )
                    
                    if not renewed:
                        logger.warning(f"Lost leadership: {self._node_id}")
                        self._state = LeadershipState.FOLLOWER
                        self._leadership = None
                        self._notify_callbacks()
                else:
                    # Try to acquire leadership
                    acquired = await self._store.try_acquire_leadership(
                        self._election_name,
                        self._node_id,
                        self._config.lease_duration_seconds,
                    )
                    
                    if acquired:
                        logger.info(f"Acquired leadership: {self._node_id}")
                        self._state = LeadershipState.LEADER
                        
                        leader_info = await self._store.get_leader(
                            self._election_name
                        )
                        
                        self._leadership = Leadership(
                            is_leader=True,
                            leader_id=self._node_id,
                            term=leader_info.term if leader_info else 0,
                            since=datetime.now(),
                        )
                        
                        self._notify_callbacks()
                
                # Sleep for heartbeat interval
                await asyncio.sleep(
                    self._config.heartbeat_interval_ms / 1000
                )
            
            except asyncio.CancelledError:
                break
            
            except Exception as e:
                logger.error(f"Election error: {e}")
                await asyncio.sleep(1)
    
    def on_leadership_change(
        self,
        callback: Callable[[Leadership], None],
    ) -> Callable[[], None]:
        """Register callback for leadership changes."""
        self._callbacks.append(callback)
        
        def unregister():
            self._callbacks.remove(callback)
        
        return unregister
    
    def _notify_callbacks(self) -> None:
        """Notify registered callbacks."""
        leadership = self._leadership or Leadership(is_leader=False)
        
        for callback in self._callbacks:
            try:
                if asyncio.iscoroutinefunction(callback):
                    asyncio.create_task(callback(leadership))
                else:
                    callback(leadership)
            except Exception as e:
                logger.error(f"Callback error: {e}")
    
    @asynccontextmanager
    async def campaign(self):
        """
        Context manager for leader campaign.
        
        Example:
            async with elector.campaign() as leadership:
                if leadership.is_leader:
                    await do_leader_work()
        """
        await self.start()
        
        try:
            # Wait briefly for initial election
            await asyncio.sleep(0.1)
            
            yield self._leadership or Leadership(is_leader=False)
        
        finally:
            await self.stop()
    
    async def wait_for_leadership(
        self,
        timeout: Optional[float] = None,
    ) -> bool:
        """Wait until this node becomes leader."""
        start = time.time()
        
        while True:
            if self.is_leader:
                return True
            
            if timeout and (time.time() - start) > timeout:
                return False
            
            await asyncio.sleep(0.1)


# Distributed Lock
class LockStore(ABC):
    """Abstract lock store."""
    
    @abstractmethod
    async def acquire(
        self,
        resource_id: str,
        holder_id: str,
        ttl_seconds: int,
    ) -> bool:
        """Acquire lock."""
        pass
    
    @abstractmethod
    async def release(
        self,
        resource_id: str,
        holder_id: str,
    ) -> bool:
        """Release lock."""
        pass
    
    @abstractmethod
    async def extend(
        self,
        resource_id: str,
        holder_id: str,
        ttl_seconds: int,
    ) -> bool:
        """Extend lock TTL."""
        pass
    
    @abstractmethod
    async def get_lock_info(
        self,
        resource_id: str,
    ) -> Optional[LockInfo]:
        """Get lock information."""
        pass


class InMemoryLockStore(LockStore):
    """In-memory lock store."""
    
    def __init__(self):
        self._locks: Dict[str, LockInfo] = {}
        self._lock = asyncio.Lock()
    
    async def acquire(
        self,
        resource_id: str,
        holder_id: str,
        ttl_seconds: int,
    ) -> bool:
        async with self._lock:
            current = self._locks.get(resource_id)
            
            # Check if lock is held by another
            if current:
                if current.expires_at > datetime.now():
                    return current.holder_id == holder_id
            
            # Acquire lock
            now = datetime.now()
            self._locks[resource_id] = LockInfo(
                lock_id=str(uuid.uuid4()),
                resource_id=resource_id,
                holder_id=holder_id,
                acquired_at=now,
                ttl_seconds=ttl_seconds,
                expires_at=now + timedelta(seconds=ttl_seconds),
            )
            
            return True
    
    async def release(
        self,
        resource_id: str,
        holder_id: str,
    ) -> bool:
        async with self._lock:
            current = self._locks.get(resource_id)
            
            if current and current.holder_id == holder_id:
                del self._locks[resource_id]
                return True
            
            return False
    
    async def extend(
        self,
        resource_id: str,
        holder_id: str,
        ttl_seconds: int,
    ) -> bool:
        async with self._lock:
            current = self._locks.get(resource_id)
            
            if current and current.holder_id == holder_id:
                current.expires_at = datetime.now() + timedelta(seconds=ttl_seconds)
                current.ttl_seconds = ttl_seconds
                return True
            
            return False
    
    async def get_lock_info(
        self,
        resource_id: str,
    ) -> Optional[LockInfo]:
        return self._locks.get(resource_id)


class DistributedLock:
    """
    Distributed lock for exclusive resource access.
    """
    
    def __init__(
        self,
        resource_id: str,
        store: LockStore,
        holder_id: Optional[str] = None,
        ttl_seconds: int = 30,
        retry_interval_ms: int = 100,
        max_retries: int = 100,
    ):
        self._resource_id = resource_id
        self._store = store
        self._holder_id = holder_id or f"{os.getpid()}-{uuid.uuid4().hex[:8]}"
        self._ttl_seconds = ttl_seconds
        self._retry_interval = retry_interval_ms / 1000
        self._max_retries = max_retries
        self._held = False
        self._extend_task: Optional[asyncio.Task] = None
    
    @property
    def resource_id(self) -> str:
        return self._resource_id
    
    @property
    def is_held(self) -> bool:
        return self._held
    
    async def try_acquire(self) -> bool:
        """Try to acquire lock without blocking."""
        acquired = await self._store.acquire(
            self._resource_id,
            self._holder_id,
            self._ttl_seconds,
        )
        
        if acquired:
            self._held = True
            self._start_extend_task()
        
        return acquired
    
    async def acquire_with_timeout(
        self,
        timeout_seconds: float,
    ) -> bool:
        """Acquire lock with timeout."""
        start = time.time()
        attempts = 0
        
        while True:
            if await self.try_acquire():
                return True
            
            if time.time() - start > timeout_seconds:
                return False
            
            attempts += 1
            if attempts >= self._max_retries:
                return False
            
            await asyncio.sleep(self._retry_interval)
    
    async def release(self) -> bool:
        """Release the lock."""
        self._stop_extend_task()
        
        if not self._held:
            return False
        
        released = await self._store.release(
            self._resource_id,
            self._holder_id,
        )
        
        self._held = not released
        
        return released
    
    def _start_extend_task(self) -> None:
        """Start automatic lock extension."""
        async def extend_loop():
            interval = self._ttl_seconds / 3
            
            while self._held:
                try:
                    await asyncio.sleep(interval)
                    
                    if self._held:
                        await self._store.extend(
                            self._resource_id,
                            self._holder_id,
                            self._ttl_seconds,
                        )
                
                except asyncio.CancelledError:
                    break
                except Exception as e:
                    logger.error(f"Lock extend error: {e}")
        
        self._extend_task = asyncio.create_task(extend_loop())
    
    def _stop_extend_task(self) -> None:
        """Stop automatic lock extension."""
        if self._extend_task:
            self._extend_task.cancel()
            self._extend_task = None
    
    @asynccontextmanager
    async def acquire(self, timeout: Optional[float] = None):
        """
        Context manager for lock acquisition.
        
        Example:
            async with lock.acquire():
                # Exclusive access
                ...
        """
        if timeout:
            acquired = await self.acquire_with_timeout(timeout)
        else:
            # Keep trying
            while not await self.try_acquire():
                await asyncio.sleep(self._retry_interval)
            acquired = True
        
        if not acquired:
            raise LockAcquisitionError(
                f"Failed to acquire lock: {self._resource_id}"
            )
        
        try:
            yield self
        finally:
            await self.release()


class Semaphore:
    """
    Distributed semaphore for limiting concurrent access.
    """
    
    def __init__(
        self,
        resource_id: str,
        max_permits: int,
        store: LockStore,
        ttl_seconds: int = 30,
    ):
        self._resource_id = resource_id
        self._max_permits = max_permits
        self._store = store
        self._ttl_seconds = ttl_seconds
        self._holder_id = f"{os.getpid()}-{uuid.uuid4().hex[:8]}"
        self._acquired_slots: List[str] = []
    
    async def acquire(self, permits: int = 1) -> bool:
        """Acquire permits."""
        for i in range(permits):
            slot = f"{self._resource_id}:{i}"
            
            for attempt in range(self._max_permits):
                slot = f"{self._resource_id}:{(i + attempt) % self._max_permits}"
                
                acquired = await self._store.acquire(
                    slot,
                    self._holder_id,
                    self._ttl_seconds,
                )
                
                if acquired:
                    self._acquired_slots.append(slot)
                    break
            else:
                # Couldn't acquire, release what we got
                await self.release()
                return False
        
        return True
    
    async def release(self) -> None:
        """Release all acquired permits."""
        for slot in self._acquired_slots:
            await self._store.release(slot, self._holder_id)
        
        self._acquired_slots.clear()


# Decorators
def leader_only(
    elector: LeaderElector,
) -> Callable:
    """
    Decorator to run function only on leader.
    
    Example:
        @leader_only(elector)
        async def leader_task():
            ...
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not elector.is_leader:
                logger.debug(f"Skipping {func.__name__}: not leader")
                return None
            
            return await func(*args, **kwargs)
        
        return wrapper
    
    return decorator


def with_lock(
    lock: DistributedLock,
    timeout: Optional[float] = None,
) -> Callable:
    """
    Decorator to acquire lock before function.
    
    Example:
        @with_lock(lock, timeout=10.0)
        async def exclusive_operation():
            ...
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            async with lock.acquire(timeout):
                return await func(*args, **kwargs)
        
        return wrapper
    
    return decorator


def coordinated(
    resource_id: str,
    store: Optional[LockStore] = None,
) -> Callable:
    """
    Decorator for coordinated (locked) execution.
    
    Example:
        @coordinated("shared-resource")
        async def update_resource():
            ...
    """
    def decorator(func: Callable) -> Callable:
        _store = store or InMemoryLockStore()
        _lock = DistributedLock(resource_id, _store)
        
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            async with _lock.acquire():
                return await func(*args, **kwargs)
        
        return wrapper
    
    return decorator


# Factory functions
def create_leader_elector(
    election_name: str,
    store: Optional[LeaderStore] = None,
    node_id: Optional[str] = None,
) -> LeaderElector:
    """Create a leader elector."""
    s = store or InMemoryLeaderStore()
    return LeaderElector(election_name, s, node_id)


def create_distributed_lock(
    resource_id: str,
    store: Optional[LockStore] = None,
    ttl_seconds: int = 30,
) -> DistributedLock:
    """Create a distributed lock."""
    s = store or InMemoryLockStore()
    return DistributedLock(resource_id, s, ttl_seconds=ttl_seconds)


def create_semaphore(
    resource_id: str,
    max_permits: int,
    store: Optional[LockStore] = None,
) -> Semaphore:
    """Create a distributed semaphore."""
    s = store or InMemoryLockStore()
    return Semaphore(resource_id, max_permits, s)


__all__ = [
    # Exceptions
    "LeaderError",
    "LockError",
    "LockAcquisitionError",
    # Enums
    "LeadershipState",
    "LockState",
    # Data classes
    "LeaderInfo",
    "Leadership",
    "LockInfo",
    "ElectionConfig",
    # Stores
    "LeaderStore",
    "InMemoryLeaderStore",
    "LockStore",
    "InMemoryLockStore",
    # Core classes
    "LeaderElector",
    "DistributedLock",
    "Semaphore",
    # Decorators
    "leader_only",
    "with_lock",
    "coordinated",
    # Factory functions
    "create_leader_elector",
    "create_distributed_lock",
    "create_semaphore",
]



"""
Agent Manager for managing multiple agents.
"""

import logging
from typing import Any, Dict, List, Optional

from .agent import Agent

logger = logging.getLogger(__name__)


class AgentManager:
    """
    Manages multiple agents with enhanced monitoring and coordination.
    
    Features:
    - Agent lifecycle management
    - Performance monitoring across agents
    - Context coordination
    - Health checks
    """
    
    def __init__(self):
        """Initialize the agent manager."""
        self.agents: Dict[str, Agent] = {}
        self.manager_metrics = {
            'total_agents_registered': 0,
            'total_agents_removed': 0,
            'total_broadcasts': 0
        }

    def register_agent(self, agent: Agent) -> None:
        """Register an agent with the manager."""
        self.agents[agent.id] = agent
        self.manager_metrics['total_agents_registered'] += 1
        logger.info("Registered agent %s with ID %s", agent.name, agent.id)

    def get_agent(self, agent_id: str) -> Optional[Agent]:
        """Get an agent by ID."""
        return self.agents.get(agent_id)

    def list_agents(self) -> List[Agent]:
        """List all registered agents."""
        return list(self.agents.values())

    def remove_agent(self, agent_id: str) -> None:
        """Remove an agent by ID."""
        if agent_id in self.agents:
            agent = self.agents[agent_id]
            agent.stop()
            del self.agents[agent_id]
            self.manager_metrics['total_agents_removed'] += 1
            logger.info("Removed agent with ID %s", agent_id)

    def broadcast(self, message: str, importance: float = 0.5) -> None:
        """
        Broadcast a message to all agents.
        
        Args:
            message: Message to broadcast
            importance: Importance score for context management
        """
        self.manager_metrics['total_broadcasts'] += 1
        for agent in self.agents.values():
            agent.log(f"Broadcast message: {message}")
            agent.add_context(f"Broadcast: {message}", importance=importance)
    
    def get_agent_by_name(self, name: str) -> Optional[Agent]:
        """Get an agent by name."""
        for agent in self.agents.values():
            if agent.name == name:
                return agent
        return None
    
    def get_agents_by_capability(self, capability: str) -> List[Agent]:
        """Get all agents with a specific capability."""
        return [
            agent for agent in self.agents.values()
            if capability in agent.capabilities
        ]
    
    def get_active_agents(self) -> List[Agent]:
        """Get all running agents."""
        return [
            agent for agent in self.agents.values()
            if agent.status == "running"
        ]
    
    def health_check(self) -> Dict[str, Any]:
        """
        Perform health check on all agents.
        
        Returns:
            Dict with health status for each agent
        """
        health_status = {}
        
        for agent_id, agent in self.agents.items():
            metrics = agent.get_performance_metrics()
            context_stats = agent.get_context_stats()
            
            health_status[agent_id] = {
                'name': agent.name,
                'status': agent.status,
                'success_rate': metrics['success_rate'],
                'total_tasks': metrics['total_tasks'],
                'error_count': metrics['error_count'],
                'context_utilization': context_stats['utilization']
            }
        
        return health_status
    
    def get_aggregate_metrics(self) -> Dict[str, Any]:
        """Get aggregated metrics across all agents."""
        total_tasks = 0
        total_successful = 0
        total_failed = 0
        total_errors = 0
        
        for agent in self.agents.values():
            metrics = agent.get_performance_metrics()
            total_tasks += metrics['total_tasks']
            total_successful += metrics['successful_tasks']
            total_failed += metrics['failed_tasks']
            total_errors += metrics['error_count']
        
        return {
            'total_agents': len(self.agents),
            'active_agents': len(self.get_active_agents()),
            'total_tasks': total_tasks,
            'total_successful': total_successful,
            'total_failed': total_failed,
            'total_errors': total_errors,
            'overall_success_rate': total_successful / total_tasks if total_tasks > 0 else 0.0,
            **self.manager_metrics
        }
    
    def stop_all_agents(self) -> None:
        """Stop all agents."""
        for agent in self.agents.values():
            agent.stop()
        logger.info("Stopped all %d agents", len(self.agents))



"""
Enterprise Document Converter Module.

Provides document format conversion between
DOCX, PDF, HTML, Markdown, and other formats.

Example:
    # Create document converter
    docs = create_document_converter()
    
    # Convert Markdown to HTML
    html = await docs.md_to_html(markdown_content)
    
    # Convert HTML to PDF
    pdf = await docs.html_to_pdf(html_content)
    
    # Convert DOCX to PDF
    pdf = await docs.convert(docx_bytes, from_format="docx", to_format="pdf")
    
    # Batch conversion
    results = await docs.batch_convert(files, to_format="pdf")
"""

from __future__ import annotations

import asyncio
import base64
import functools
import io
import json
import logging
import re
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class ConversionError(Exception):
    """Document conversion error."""
    pass


class FormatError(ConversionError):
    """Format error."""
    pass


class UnsupportedFormatError(ConversionError):
    """Unsupported format error."""
    pass


class DocumentFormat(str, Enum):
    """Document formats."""
    # Text formats
    TXT = "txt"
    MARKDOWN = "md"
    HTML = "html"
    RTF = "rtf"
    # Office formats
    DOCX = "docx"
    DOC = "doc"
    ODT = "odt"
    # Presentation
    PPTX = "pptx"
    PPT = "ppt"
    ODP = "odp"
    # Spreadsheet
    XLSX = "xlsx"
    XLS = "xls"
    CSV = "csv"
    ODS = "ods"
    # PDF
    PDF = "pdf"
    # Other
    EPUB = "epub"
    LATEX = "tex"
    RST = "rst"
    JSON = "json"
    XML = "xml"


@dataclass
class DocumentInfo:
    """Document information."""
    format: DocumentFormat
    size_bytes: int = 0
    page_count: int = 0
    word_count: int = 0
    char_count: int = 0
    title: str = ""
    author: str = ""
    created_at: Optional[datetime] = None
    modified_at: Optional[datetime] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ConversionOptions:
    """Conversion options."""
    from_format: Optional[DocumentFormat] = None
    to_format: DocumentFormat = DocumentFormat.PDF
    page_size: str = "A4"
    orientation: str = "portrait"
    margin_top: float = 1.0  # inches
    margin_bottom: float = 1.0
    margin_left: float = 1.0
    margin_right: float = 1.0
    header: Optional[str] = None
    footer: Optional[str] = None
    font_family: str = "Arial"
    font_size: int = 12
    embed_images: bool = True
    table_of_contents: bool = False
    syntax_highlighting: bool = True


@dataclass
class ConvertedDocument:
    """Converted document result."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    content: bytes = b""
    format: DocumentFormat = DocumentFormat.PDF
    original_format: Optional[DocumentFormat] = None
    size_bytes: int = 0
    page_count: int = 0
    converted_at: datetime = field(default_factory=datetime.utcnow)


# Format converters
class FormatConverter(ABC):
    """Abstract format converter."""
    
    @property
    @abstractmethod
    def supported_from(self) -> List[DocumentFormat]:
        """Formats this converter can read."""
        pass
    
    @property
    @abstractmethod
    def supported_to(self) -> List[DocumentFormat]:
        """Formats this converter can write."""
        pass
    
    @abstractmethod
    async def convert(
        self,
        data: bytes,
        from_format: DocumentFormat,
        to_format: DocumentFormat,
        options: ConversionOptions,
    ) -> bytes:
        """Convert document."""
        pass


class MarkdownConverter(FormatConverter):
    """Markdown converter."""
    
    @property
    def supported_from(self) -> List[DocumentFormat]:
        return [DocumentFormat.MARKDOWN]
    
    @property
    def supported_to(self) -> List[DocumentFormat]:
        return [DocumentFormat.HTML]
    
    async def convert(
        self,
        data: bytes,
        from_format: DocumentFormat,
        to_format: DocumentFormat,
        options: ConversionOptions,
    ) -> bytes:
        """Convert Markdown to HTML."""
        text = data.decode()
        
        # Basic Markdown to HTML conversion
        html = self._markdown_to_html(text, options)
        
        return html.encode()
    
    def _markdown_to_html(
        self,
        text: str,
        options: ConversionOptions,
    ) -> str:
        """Convert Markdown to HTML."""
        lines = text.split('\n')
        html_lines = []
        in_code_block = False
        in_list = False
        list_type = None
        
        for line in lines:
            # Code blocks
            if line.startswith('```'):
                if in_code_block:
                    html_lines.append('</code></pre>')
                    in_code_block = False
                else:
                    lang = line[3:].strip()
                    html_lines.append(f'<pre><code class="language-{lang}">')
                    in_code_block = True
                continue
            
            if in_code_block:
                html_lines.append(self._escape_html(line))
                continue
            
            # Headers
            if line.startswith('#'):
                level = len(line.split()[0])
                text_content = line[level:].strip()
                html_lines.append(f'<h{level}>{text_content}</h{level}>')
                continue
            
            # Bold and italic
            line = re.sub(r'\*\*(.+?)\*\*', r'<strong>\1</strong>', line)
            line = re.sub(r'\*(.+?)\*', r'<em>\1</em>', line)
            line = re.sub(r'__(.+?)__', r'<strong>\1</strong>', line)
            line = re.sub(r'_(.+?)_', r'<em>\1</em>', line)
            
            # Links
            line = re.sub(
                r'\[(.+?)\]\((.+?)\)',
                r'<a href="\2">\1</a>',
                line,
            )
            
            # Images
            line = re.sub(
                r'!\[(.+?)\]\((.+?)\)',
                r'<img src="\2" alt="\1">',
                line,
            )
            
            # Inline code
            line = re.sub(r'`(.+?)`', r'<code>\1</code>', line)
            
            # Lists
            if line.startswith('- ') or line.startswith('* '):
                if not in_list or list_type != 'ul':
                    if in_list:
                        html_lines.append(f'</{list_type}>')
                    html_lines.append('<ul>')
                    in_list = True
                    list_type = 'ul'
                html_lines.append(f'<li>{line[2:]}</li>')
                continue
            
            if re.match(r'^\d+\. ', line):
                if not in_list or list_type != 'ol':
                    if in_list:
                        html_lines.append(f'</{list_type}>')
                    html_lines.append('<ol>')
                    in_list = True
                    list_type = 'ol'
                text_content = re.sub(r'^\d+\. ', '', line)
                html_lines.append(f'<li>{text_content}</li>')
                continue
            
            # Close list if needed
            if in_list and line.strip() == '':
                html_lines.append(f'</{list_type}>')
                in_list = False
                list_type = None
            
            # Horizontal rule
            if line.strip() in ('---', '***', '___'):
                html_lines.append('<hr>')
                continue
            
            # Blockquote
            if line.startswith('> '):
                html_lines.append(f'<blockquote>{line[2:]}</blockquote>')
                continue
            
            # Paragraph
            if line.strip():
                html_lines.append(f'<p>{line}</p>')
            else:
                html_lines.append('')
        
        # Close any open list
        if in_list:
            html_lines.append(f'</{list_type}>')
        
        # Wrap in HTML document
        html_content = '\n'.join(html_lines)
        
        return f'''<!DOCTYPE html>
<html>
<head>
    <meta charset="UTF-8">
    <style>
        body {{ font-family: {options.font_family}; font-size: {options.font_size}px; margin: 40px; }}
        pre {{ background: #f4f4f4; padding: 10px; border-radius: 5px; }}
        code {{ background: #f4f4f4; padding: 2px 5px; border-radius: 3px; }}
        blockquote {{ border-left: 3px solid #ccc; padding-left: 15px; color: #666; }}
    </style>
</head>
<body>
{html_content}
</body>
</html>'''
    
    def _escape_html(self, text: str) -> str:
        """Escape HTML characters."""
        return (
            text.replace('&', '&amp;')
            .replace('<', '&lt;')
            .replace('>', '&gt;')
        )


class HTMLConverter(FormatConverter):
    """HTML converter."""
    
    @property
    def supported_from(self) -> List[DocumentFormat]:
        return [DocumentFormat.HTML]
    
    @property
    def supported_to(self) -> List[DocumentFormat]:
        return [DocumentFormat.TXT, DocumentFormat.MARKDOWN]
    
    async def convert(
        self,
        data: bytes,
        from_format: DocumentFormat,
        to_format: DocumentFormat,
        options: ConversionOptions,
    ) -> bytes:
        """Convert HTML."""
        text = data.decode()
        
        if to_format == DocumentFormat.TXT:
            result = self._html_to_text(text)
        elif to_format == DocumentFormat.MARKDOWN:
            result = self._html_to_markdown(text)
        else:
            raise UnsupportedFormatError(f"Cannot convert to {to_format}")
        
        return result.encode()
    
    def _html_to_text(self, html: str) -> str:
        """Convert HTML to plain text."""
        # Remove script and style
        text = re.sub(r'<script[^>]*>.*?</script>', '', html, flags=re.DOTALL)
        text = re.sub(r'<style[^>]*>.*?</style>', '', text, flags=re.DOTALL)
        
        # Replace common elements
        text = re.sub(r'<br\s*/?>', '\n', text)
        text = re.sub(r'</p>', '\n\n', text)
        text = re.sub(r'</div>', '\n', text)
        text = re.sub(r'</h\d>', '\n\n', text)
        text = re.sub(r'</li>', '\n', text)
        
        # Remove tags
        text = re.sub(r'<[^>]+>', '', text)
        
        # Decode entities
        text = text.replace('&nbsp;', ' ')
        text = text.replace('&amp;', '&')
        text = text.replace('&lt;', '<')
        text = text.replace('&gt;', '>')
        text = text.replace('&quot;', '"')
        
        # Clean up whitespace
        text = re.sub(r'\n\s*\n', '\n\n', text)
        text = text.strip()
        
        return text
    
    def _html_to_markdown(self, html: str) -> str:
        """Convert HTML to Markdown."""
        text = html
        
        # Headers
        for i in range(1, 7):
            text = re.sub(
                rf'<h{i}[^>]*>(.*?)</h{i}>',
                rf'{"#" * i} \1\n',
                text,
                flags=re.DOTALL,
            )
        
        # Bold and italic
        text = re.sub(r'<strong[^>]*>(.*?)</strong>', r'**\1**', text)
        text = re.sub(r'<b[^>]*>(.*?)</b>', r'**\1**', text)
        text = re.sub(r'<em[^>]*>(.*?)</em>', r'*\1*', text)
        text = re.sub(r'<i[^>]*>(.*?)</i>', r'*\1*', text)
        
        # Links
        text = re.sub(
            r'<a[^>]*href=["\']([^"\']+)["\'][^>]*>(.*?)</a>',
            r'[\2](\1)',
            text,
        )
        
        # Images
        text = re.sub(
            r'<img[^>]*src=["\']([^"\']+)["\'][^>]*alt=["\']([^"\']*)["\'][^>]*/?>',
            r'![\2](\1)',
            text,
        )
        
        # Lists
        text = re.sub(r'<li[^>]*>(.*?)</li>', r'- \1\n', text, flags=re.DOTALL)
        text = re.sub(r'</?[uo]l[^>]*>', '', text)
        
        # Code
        text = re.sub(r'<code[^>]*>(.*?)</code>', r'`\1`', text)
        text = re.sub(
            r'<pre[^>]*><code[^>]*>(.*?)</code></pre>',
            r'```\n\1\n```',
            text,
            flags=re.DOTALL,
        )
        
        # Blockquote
        text = re.sub(
            r'<blockquote[^>]*>(.*?)</blockquote>',
            lambda m: '> ' + m.group(1).replace('\n', '\n> '),
            text,
            flags=re.DOTALL,
        )
        
        # Paragraphs and line breaks
        text = re.sub(r'<br\s*/?>', '\n', text)
        text = re.sub(r'<p[^>]*>(.*?)</p>', r'\1\n\n', text, flags=re.DOTALL)
        
        # Remove remaining tags
        text = re.sub(r'<[^>]+>', '', text)
        
        # Decode entities
        text = text.replace('&nbsp;', ' ')
        text = text.replace('&amp;', '&')
        text = text.replace('&lt;', '<')
        text = text.replace('&gt;', '>')
        
        # Clean up
        text = re.sub(r'\n{3,}', '\n\n', text)
        text = text.strip()
        
        return text


class MockPDFConverter(FormatConverter):
    """Mock PDF converter for testing."""
    
    @property
    def supported_from(self) -> List[DocumentFormat]:
        return [DocumentFormat.HTML, DocumentFormat.MARKDOWN, DocumentFormat.TXT]
    
    @property
    def supported_to(self) -> List[DocumentFormat]:
        return [DocumentFormat.PDF]
    
    async def convert(
        self,
        data: bytes,
        from_format: DocumentFormat,
        to_format: DocumentFormat,
        options: ConversionOptions,
    ) -> bytes:
        """Mock PDF conversion."""
        info = {
            "conversion": f"{from_format.value} -> {to_format.value}",
            "page_size": options.page_size,
            "orientation": options.orientation,
            "original_size": len(data),
            "converted_at": datetime.utcnow().isoformat(),
        }
        return json.dumps(info, indent=2).encode()


class DocumentConverter:
    """
    Document conversion service.
    """
    
    def __init__(
        self,
        converters: Optional[List[FormatConverter]] = None,
    ):
        self._converters = converters or [
            MarkdownConverter(),
            HTMLConverter(),
            MockPDFConverter(),
        ]
    
    def _get_converter(
        self,
        from_format: DocumentFormat,
        to_format: DocumentFormat,
    ) -> Optional[FormatConverter]:
        """Find converter for formats."""
        for converter in self._converters:
            if (
                from_format in converter.supported_from
                and to_format in converter.supported_to
            ):
                return converter
        return None
    
    def can_convert(
        self,
        from_format: DocumentFormat,
        to_format: DocumentFormat,
    ) -> bool:
        """Check if conversion is supported."""
        return self._get_converter(from_format, to_format) is not None
    
    def get_supported_conversions(self) -> List[Tuple[DocumentFormat, DocumentFormat]]:
        """Get list of supported conversions."""
        conversions = []
        for converter in self._converters:
            for from_fmt in converter.supported_from:
                for to_fmt in converter.supported_to:
                    conversions.append((from_fmt, to_fmt))
        return conversions
    
    async def convert(
        self,
        data: bytes,
        from_format: Optional[Union[str, DocumentFormat]] = None,
        to_format: Union[str, DocumentFormat] = DocumentFormat.PDF,
        options: Optional[ConversionOptions] = None,
    ) -> ConvertedDocument:
        """
        Convert document.
        
        Args:
            data: Document bytes
            from_format: Source format (auto-detect if None)
            to_format: Target format
            options: Conversion options
            
        Returns:
            Converted document
        """
        # Normalize formats
        if isinstance(from_format, str):
            from_format = DocumentFormat(from_format)
        if isinstance(to_format, str):
            to_format = DocumentFormat(to_format)
        
        # Auto-detect format if not provided
        if from_format is None:
            from_format = self._detect_format(data)
        
        options = options or ConversionOptions(
            from_format=from_format,
            to_format=to_format,
        )
        
        # Find converter
        converter = self._get_converter(from_format, to_format)
        
        if not converter:
            # Try multi-step conversion
            content = await self._multi_step_convert(
                data, from_format, to_format, options
            )
        else:
            content = await converter.convert(
                data, from_format, to_format, options
            )
        
        return ConvertedDocument(
            content=content,
            format=to_format,
            original_format=from_format,
            size_bytes=len(content),
        )
    
    async def _multi_step_convert(
        self,
        data: bytes,
        from_format: DocumentFormat,
        to_format: DocumentFormat,
        options: ConversionOptions,
    ) -> bytes:
        """Try multi-step conversion."""
        # Try converting through HTML
        if from_format == DocumentFormat.MARKDOWN and to_format == DocumentFormat.PDF:
            # MD -> HTML -> PDF
            html = await self.md_to_html(data.decode())
            return (await self.html_to_pdf(html)).content
        
        raise UnsupportedFormatError(
            f"Cannot convert from {from_format} to {to_format}"
        )
    
    def _detect_format(self, data: bytes) -> DocumentFormat:
        """Detect document format."""
        # Try to decode as text
        try:
            text = data.decode()
            
            # Check for HTML
            if '<html' in text.lower() or '<!doctype html' in text.lower():
                return DocumentFormat.HTML
            
            # Check for Markdown indicators
            if re.search(r'^#{1,6}\s', text, re.MULTILINE):
                return DocumentFormat.MARKDOWN
            
            return DocumentFormat.TXT
            
        except:
            pass
        
        # Check magic bytes
        if data.startswith(b'%PDF'):
            return DocumentFormat.PDF
        
        if data.startswith(b'PK'):
            # Could be DOCX, XLSX, PPTX
            return DocumentFormat.DOCX
        
        return DocumentFormat.TXT
    
    # Convenience methods
    async def md_to_html(
        self,
        markdown: str,
        options: Optional[ConversionOptions] = None,
    ) -> str:
        """Convert Markdown to HTML."""
        result = await self.convert(
            markdown.encode(),
            from_format=DocumentFormat.MARKDOWN,
            to_format=DocumentFormat.HTML,
            options=options,
        )
        return result.content.decode()
    
    async def html_to_pdf(
        self,
        html: str,
        options: Optional[ConversionOptions] = None,
    ) -> ConvertedDocument:
        """Convert HTML to PDF."""
        return await self.convert(
            html.encode(),
            from_format=DocumentFormat.HTML,
            to_format=DocumentFormat.PDF,
            options=options,
        )
    
    async def html_to_text(
        self,
        html: str,
    ) -> str:
        """Convert HTML to plain text."""
        result = await self.convert(
            html.encode(),
            from_format=DocumentFormat.HTML,
            to_format=DocumentFormat.TXT,
        )
        return result.content.decode()
    
    async def html_to_md(
        self,
        html: str,
    ) -> str:
        """Convert HTML to Markdown."""
        result = await self.convert(
            html.encode(),
            from_format=DocumentFormat.HTML,
            to_format=DocumentFormat.MARKDOWN,
        )
        return result.content.decode()
    
    async def md_to_pdf(
        self,
        markdown: str,
        options: Optional[ConversionOptions] = None,
    ) -> ConvertedDocument:
        """Convert Markdown to PDF."""
        # First convert to HTML
        html = await self.md_to_html(markdown, options)
        
        # Then convert to PDF
        return await self.html_to_pdf(html, options)
    
    async def batch_convert(
        self,
        files: List[Tuple[str, bytes]],
        to_format: DocumentFormat,
        options: Optional[ConversionOptions] = None,
    ) -> List[ConvertedDocument]:
        """
        Batch convert files.
        
        Args:
            files: List of (filename, content) tuples
            to_format: Target format
            options: Conversion options
            
        Returns:
            List of converted documents
        """
        tasks = []
        
        for filename, content in files:
            # Detect format from extension
            ext = Path(filename).suffix.lstrip('.')
            try:
                from_format = DocumentFormat(ext)
            except ValueError:
                from_format = None
            
            tasks.append(
                self.convert(content, from_format, to_format, options)
            )
        
        return await asyncio.gather(*tasks, return_exceptions=True)
    
    async def save(
        self,
        document: ConvertedDocument,
        path: Union[str, Path],
    ) -> Path:
        """Save converted document."""
        path = Path(path)
        path.write_bytes(document.content)
        return path


# Factory functions
def create_document_converter(
    converters: Optional[List[FormatConverter]] = None,
) -> DocumentConverter:
    """Create document converter."""
    return DocumentConverter(converters)


def create_conversion_options(
    to_format: DocumentFormat = DocumentFormat.PDF,
    page_size: str = "A4",
    **kwargs,
) -> ConversionOptions:
    """Create conversion options."""
    return ConversionOptions(to_format=to_format, page_size=page_size, **kwargs)


__all__ = [
    # Exceptions
    "ConversionError",
    "FormatError",
    "UnsupportedFormatError",
    # Enums
    "DocumentFormat",
    # Data classes
    "DocumentInfo",
    "ConversionOptions",
    "ConvertedDocument",
    # Converters
    "FormatConverter",
    "MarkdownConverter",
    "HTMLConverter",
    "MockPDFConverter",
    # Service
    "DocumentConverter",
    # Factory functions
    "create_document_converter",
    "create_conversion_options",
]



"""
Enterprise Knowledge Manager Module.

Knowledge base management, documentation search,
article versioning, and content organization.

Example:
    # Create knowledge manager
    kb = create_knowledge_manager()
    
    # Add article
    article = await kb.add_article(
        title="Getting Started",
        content="Welcome to the platform...",
        category="guides",
    )
    
    # Search
    results = await kb.search("how to configure")
    
    # Get article
    article = await kb.get_article(article.id)
"""

from __future__ import annotations

import asyncio
import logging
import re
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
)

logger = logging.getLogger(__name__)


class KnowledgeError(Exception):
    """Knowledge error."""
    pass


class ArticleStatus(str, Enum):
    """Article status."""
    DRAFT = "draft"
    REVIEW = "review"
    PUBLISHED = "published"
    ARCHIVED = "archived"


class ContentType(str, Enum):
    """Content type."""
    MARKDOWN = "markdown"
    HTML = "html"
    PLAIN_TEXT = "plain_text"
    RICH_TEXT = "rich_text"


class ArticleType(str, Enum):
    """Article type."""
    ARTICLE = "article"
    FAQ = "faq"
    TUTORIAL = "tutorial"
    GUIDE = "guide"
    REFERENCE = "reference"
    TROUBLESHOOTING = "troubleshooting"


@dataclass
class Article:
    """Knowledge article."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    slug: str = ""
    title: str = ""
    content: str = ""
    summary: str = ""
    
    # Classification
    article_type: ArticleType = ArticleType.ARTICLE
    status: ArticleStatus = ArticleStatus.DRAFT
    content_type: ContentType = ContentType.MARKDOWN
    
    # Organization
    category: str = ""
    subcategory: str = ""
    tags: List[str] = field(default_factory=list)
    
    # Versioning
    version: int = 1
    parent_id: Optional[str] = None  # For version history
    
    # Authors
    author: str = ""
    contributors: List[str] = field(default_factory=list)
    
    # Metadata
    featured: bool = False
    pinned: bool = False
    view_count: int = 0
    helpful_count: int = 0
    not_helpful_count: int = 0
    
    # Related
    related_articles: List[str] = field(default_factory=list)
    
    # Dates
    created_at: datetime = field(default_factory=datetime.utcnow)
    updated_at: datetime = field(default_factory=datetime.utcnow)
    published_at: Optional[datetime] = None
    
    # Custom
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Category:
    """Knowledge category."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    slug: str = ""
    description: str = ""
    parent_id: Optional[str] = None
    icon: str = ""
    order: int = 0
    article_count: int = 0
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class SearchResult:
    """Search result."""
    article_id: str = ""
    title: str = ""
    snippet: str = ""
    score: float = 0.0
    highlights: List[str] = field(default_factory=list)


@dataclass
class ArticleVersion:
    """Article version."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    article_id: str = ""
    version: int = 1
    title: str = ""
    content: str = ""
    author: str = ""
    comment: str = ""
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class KnowledgeStats:
    """Knowledge base statistics."""
    total_articles: int = 0
    published: int = 0
    draft: int = 0
    categories: int = 0
    total_views: int = 0
    most_viewed: List[str] = field(default_factory=list)
    recently_updated: List[str] = field(default_factory=list)


# Article store
class ArticleStore(ABC):
    """Article storage."""
    
    @abstractmethod
    async def save(self, article: Article) -> None:
        pass
    
    @abstractmethod
    async def get(self, article_id: str) -> Optional[Article]:
        pass
    
    @abstractmethod
    async def get_by_slug(self, slug: str) -> Optional[Article]:
        pass
    
    @abstractmethod
    async def list_all(self) -> List[Article]:
        pass
    
    @abstractmethod
    async def delete(self, article_id: str) -> bool:
        pass
    
    @abstractmethod
    async def search(self, query: str) -> List[SearchResult]:
        pass


class InMemoryArticleStore(ArticleStore):
    """In-memory article store."""
    
    def __init__(self):
        self._articles: Dict[str, Article] = {}
        self._by_slug: Dict[str, str] = {}
    
    async def save(self, article: Article) -> None:
        self._articles[article.id] = article
        if article.slug:
            self._by_slug[article.slug] = article.id
    
    async def get(self, article_id: str) -> Optional[Article]:
        return self._articles.get(article_id)
    
    async def get_by_slug(self, slug: str) -> Optional[Article]:
        article_id = self._by_slug.get(slug)
        if article_id:
            return self._articles.get(article_id)
        return None
    
    async def list_all(self) -> List[Article]:
        return list(self._articles.values())
    
    async def delete(self, article_id: str) -> bool:
        article = self._articles.get(article_id)
        if article:
            del self._articles[article_id]
            if article.slug:
                self._by_slug.pop(article.slug, None)
            return True
        return False
    
    async def search(self, query: str) -> List[SearchResult]:
        results = []
        query_lower = query.lower()
        terms = query_lower.split()
        
        for article in self._articles.values():
            if article.status != ArticleStatus.PUBLISHED:
                continue
            
            score = 0.0
            highlights = []
            
            title_lower = article.title.lower()
            content_lower = article.content.lower()
            
            for term in terms:
                if term in title_lower:
                    score += 10.0
                    highlights.append(f"Title: ...{term}...")
                
                if term in content_lower:
                    score += 1.0
                    # Find snippet
                    idx = content_lower.find(term)
                    if idx >= 0:
                        start = max(0, idx - 30)
                        end = min(len(article.content), idx + len(term) + 30)
                        snippet = article.content[start:end]
                        highlights.append(f"...{snippet}...")
                
                for tag in article.tags:
                    if term in tag.lower():
                        score += 5.0
            
            if score > 0:
                snippet = article.summary or article.content[:200]
                results.append(SearchResult(
                    article_id=article.id,
                    title=article.title,
                    snippet=snippet,
                    score=score,
                    highlights=highlights[:3],
                ))
        
        return sorted(results, key=lambda r: r.score, reverse=True)


# Category store
class CategoryStore(ABC):
    """Category storage."""
    
    @abstractmethod
    async def save(self, category: Category) -> None:
        pass
    
    @abstractmethod
    async def get(self, category_id: str) -> Optional[Category]:
        pass
    
    @abstractmethod
    async def get_by_slug(self, slug: str) -> Optional[Category]:
        pass
    
    @abstractmethod
    async def list_all(self) -> List[Category]:
        pass


class InMemoryCategoryStore(CategoryStore):
    """In-memory category store."""
    
    def __init__(self):
        self._categories: Dict[str, Category] = {}
        self._by_slug: Dict[str, str] = {}
    
    async def save(self, category: Category) -> None:
        self._categories[category.id] = category
        if category.slug:
            self._by_slug[category.slug] = category.id
    
    async def get(self, category_id: str) -> Optional[Category]:
        return self._categories.get(category_id)
    
    async def get_by_slug(self, slug: str) -> Optional[Category]:
        cat_id = self._by_slug.get(slug)
        if cat_id:
            return self._categories.get(cat_id)
        return None
    
    async def list_all(self) -> List[Category]:
        return sorted(self._categories.values(), key=lambda c: c.order)


# Version store
class VersionStore(ABC):
    """Version storage."""
    
    @abstractmethod
    async def save(self, version: ArticleVersion) -> None:
        pass
    
    @abstractmethod
    async def list_by_article(self, article_id: str) -> List[ArticleVersion]:
        pass
    
    @abstractmethod
    async def get(self, version_id: str) -> Optional[ArticleVersion]:
        pass


class InMemoryVersionStore(VersionStore):
    """In-memory version store."""
    
    def __init__(self):
        self._versions: Dict[str, ArticleVersion] = {}
        self._by_article: Dict[str, List[str]] = {}
    
    async def save(self, version: ArticleVersion) -> None:
        self._versions[version.id] = version
        
        if version.article_id not in self._by_article:
            self._by_article[version.article_id] = []
        
        self._by_article[version.article_id].append(version.id)
    
    async def list_by_article(self, article_id: str) -> List[ArticleVersion]:
        version_ids = self._by_article.get(article_id, [])
        versions = [self._versions[vid] for vid in version_ids if vid in self._versions]
        return sorted(versions, key=lambda v: v.version, reverse=True)
    
    async def get(self, version_id: str) -> Optional[ArticleVersion]:
        return self._versions.get(version_id)


# Text utilities
class TextUtils:
    """Text utilities."""
    
    @staticmethod
    def slugify(text: str) -> str:
        """Convert text to slug."""
        text = text.lower().strip()
        text = re.sub(r"[^\w\s-]", "", text)
        text = re.sub(r"[\s_]+", "-", text)
        text = re.sub(r"-+", "-", text)
        return text.strip("-")
    
    @staticmethod
    def generate_summary(content: str, max_length: int = 200) -> str:
        """Generate summary from content."""
        # Strip markdown/html
        text = re.sub(r"[#*_`\[\]()]", "", content)
        text = re.sub(r"<[^>]+>", "", text)
        text = re.sub(r"\s+", " ", text).strip()
        
        if len(text) <= max_length:
            return text
        
        # Find sentence boundary
        truncated = text[:max_length]
        last_period = truncated.rfind(".")
        last_space = truncated.rfind(" ")
        
        if last_period > max_length * 0.5:
            return truncated[:last_period + 1]
        elif last_space > 0:
            return truncated[:last_space] + "..."
        
        return truncated + "..."
    
    @staticmethod
    def extract_keywords(content: str, max_keywords: int = 10) -> List[str]:
        """Extract keywords from content."""
        # Simple word frequency
        words = re.findall(r"\b\w{4,}\b", content.lower())
        
        # Filter common words
        stopwords = {
            "this", "that", "with", "from", "have", "been", "will",
            "would", "could", "should", "what", "when", "where", "which",
            "their", "there", "these", "those", "about", "into", "your",
        }
        
        word_freq: Dict[str, int] = {}
        for word in words:
            if word not in stopwords:
                word_freq[word] = word_freq.get(word, 0) + 1
        
        sorted_words = sorted(word_freq.items(), key=lambda x: x[1], reverse=True)
        return [word for word, _ in sorted_words[:max_keywords]]


# Knowledge manager
class KnowledgeManager:
    """Knowledge manager."""
    
    def __init__(
        self,
        article_store: Optional[ArticleStore] = None,
        category_store: Optional[CategoryStore] = None,
        version_store: Optional[VersionStore] = None,
    ):
        self._article_store = article_store or InMemoryArticleStore()
        self._category_store = category_store or InMemoryCategoryStore()
        self._version_store = version_store or InMemoryVersionStore()
        self._listeners: List[Callable] = []
    
    async def add_article(
        self,
        title: str,
        content: str,
        category: str = "",
        article_type: ArticleType = ArticleType.ARTICLE,
        author: str = "",
        tags: Optional[List[str]] = None,
        status: ArticleStatus = ArticleStatus.DRAFT,
        **kwargs,
    ) -> Article:
        """Add article."""
        slug = TextUtils.slugify(title)
        summary = TextUtils.generate_summary(content)
        
        # Ensure unique slug
        existing = await self._article_store.get_by_slug(slug)
        if existing:
            slug = f"{slug}-{uuid.uuid4().hex[:6]}"
        
        article = Article(
            slug=slug,
            title=title,
            content=content,
            summary=summary,
            category=category,
            article_type=article_type,
            author=author,
            tags=tags or [],
            status=status,
            **kwargs,
        )
        
        await self._article_store.save(article)
        
        # Save initial version
        await self._save_version(article, "Initial version")
        
        # Update category count
        if category:
            cat = await self._category_store.get_by_slug(category)
            if cat:
                cat.article_count += 1
                await self._category_store.save(cat)
        
        logger.info(f"Article added: {title}")
        
        await self._notify("add", article)
        
        return article
    
    async def get_article(
        self,
        article_id: str,
        increment_views: bool = True,
    ) -> Optional[Article]:
        """Get article by ID."""
        article = await self._article_store.get(article_id)
        
        if article and increment_views:
            article.view_count += 1
            await self._article_store.save(article)
        
        return article
    
    async def get_by_slug(self, slug: str) -> Optional[Article]:
        """Get article by slug."""
        return await self._article_store.get_by_slug(slug)
    
    async def update_article(
        self,
        article_id: str,
        author: str = "",
        comment: str = "",
        **updates,
    ) -> Optional[Article]:
        """Update article."""
        article = await self._article_store.get(article_id)
        
        if not article:
            return None
        
        for key, value in updates.items():
            if hasattr(article, key):
                setattr(article, key, value)
        
        # Update summary if content changed
        if "content" in updates:
            article.summary = TextUtils.generate_summary(article.content)
        
        article.version += 1
        article.updated_at = datetime.utcnow()
        
        if author and author not in article.contributors:
            article.contributors.append(author)
        
        await self._article_store.save(article)
        
        # Save version
        await self._save_version(article, comment or "Updated")
        
        logger.info(f"Article updated: {article.title} (v{article.version})")
        
        await self._notify("update", article)
        
        return article
    
    async def publish(self, article_id: str) -> Optional[Article]:
        """Publish article."""
        article = await self._article_store.get(article_id)
        
        if not article:
            return None
        
        article.status = ArticleStatus.PUBLISHED
        article.published_at = datetime.utcnow()
        article.updated_at = datetime.utcnow()
        
        await self._article_store.save(article)
        
        logger.info(f"Article published: {article.title}")
        
        await self._notify("publish", article)
        
        return article
    
    async def archive(self, article_id: str) -> Optional[Article]:
        """Archive article."""
        article = await self._article_store.get(article_id)
        
        if not article:
            return None
        
        article.status = ArticleStatus.ARCHIVED
        article.updated_at = datetime.utcnow()
        
        await self._article_store.save(article)
        
        logger.info(f"Article archived: {article.title}")
        
        return article
    
    async def delete_article(self, article_id: str) -> bool:
        """Delete article."""
        article = await self._article_store.get(article_id)
        
        if not article:
            return False
        
        result = await self._article_store.delete(article_id)
        
        if result:
            logger.info(f"Article deleted: {article.title}")
            await self._notify("delete", article)
        
        return result
    
    async def search(
        self,
        query: str,
        category: Optional[str] = None,
        tags: Optional[List[str]] = None,
        limit: int = 20,
    ) -> List[SearchResult]:
        """Search articles."""
        results = await self._article_store.search(query)
        
        # Filter by category
        if category:
            filtered = []
            for result in results:
                article = await self._article_store.get(result.article_id)
                if article and article.category == category:
                    filtered.append(result)
            results = filtered
        
        # Filter by tags
        if tags:
            filtered = []
            for result in results:
                article = await self._article_store.get(result.article_id)
                if article and any(t in article.tags for t in tags):
                    filtered.append(result)
            results = filtered
        
        return results[:limit]
    
    async def list_articles(
        self,
        category: Optional[str] = None,
        status: Optional[ArticleStatus] = None,
        article_type: Optional[ArticleType] = None,
        author: Optional[str] = None,
        featured: Optional[bool] = None,
    ) -> List[Article]:
        """List articles with filters."""
        articles = await self._article_store.list_all()
        
        if category:
            articles = [a for a in articles if a.category == category]
        if status:
            articles = [a for a in articles if a.status == status]
        if article_type:
            articles = [a for a in articles if a.article_type == article_type]
        if author:
            articles = [a for a in articles if a.author == author]
        if featured is not None:
            articles = [a for a in articles if a.featured == featured]
        
        return sorted(articles, key=lambda a: a.updated_at, reverse=True)
    
    async def add_category(
        self,
        name: str,
        description: str = "",
        parent_id: Optional[str] = None,
        icon: str = "",
        order: int = 0,
    ) -> Category:
        """Add category."""
        slug = TextUtils.slugify(name)
        
        category = Category(
            name=name,
            slug=slug,
            description=description,
            parent_id=parent_id,
            icon=icon,
            order=order,
        )
        
        await self._category_store.save(category)
        
        logger.info(f"Category added: {name}")
        
        return category
    
    async def get_category(self, slug: str) -> Optional[Category]:
        """Get category by slug."""
        return await self._category_store.get_by_slug(slug)
    
    async def list_categories(
        self,
        parent_id: Optional[str] = None,
    ) -> List[Category]:
        """List categories."""
        categories = await self._category_store.list_all()
        
        if parent_id is not None:
            categories = [c for c in categories if c.parent_id == parent_id]
        
        return categories
    
    async def get_versions(self, article_id: str) -> List[ArticleVersion]:
        """Get article versions."""
        return await self._version_store.list_by_article(article_id)
    
    async def restore_version(
        self,
        article_id: str,
        version_id: str,
        author: str = "",
    ) -> Optional[Article]:
        """Restore article to specific version."""
        version = await self._version_store.get(version_id)
        
        if not version or version.article_id != article_id:
            return None
        
        return await self.update_article(
            article_id,
            title=version.title,
            content=version.content,
            author=author,
            comment=f"Restored from version {version.version}",
        )
    
    async def mark_helpful(
        self,
        article_id: str,
        helpful: bool,
    ) -> Optional[Article]:
        """Mark article as helpful/not helpful."""
        article = await self._article_store.get(article_id)
        
        if not article:
            return None
        
        if helpful:
            article.helpful_count += 1
        else:
            article.not_helpful_count += 1
        
        await self._article_store.save(article)
        
        return article
    
    async def get_related(
        self,
        article_id: str,
        limit: int = 5,
    ) -> List[Article]:
        """Get related articles."""
        article = await self._article_store.get(article_id)
        
        if not article:
            return []
        
        # Get explicitly related
        related = []
        for rel_id in article.related_articles:
            rel = await self._article_store.get(rel_id)
            if rel and rel.status == ArticleStatus.PUBLISHED:
                related.append(rel)
        
        # Find similar by category/tags
        if len(related) < limit:
            all_articles = await self._article_store.list_all()
            
            for art in all_articles:
                if art.id == article.id or art in related:
                    continue
                if art.status != ArticleStatus.PUBLISHED:
                    continue
                
                # Same category or shared tags
                if art.category == article.category:
                    related.append(art)
                elif any(t in article.tags for t in art.tags):
                    related.append(art)
                
                if len(related) >= limit:
                    break
        
        return related[:limit]
    
    async def get_stats(self) -> KnowledgeStats:
        """Get knowledge base statistics."""
        articles = await self._article_store.list_all()
        categories = await self._category_store.list_all()
        
        stats = KnowledgeStats(
            total_articles=len(articles),
            categories=len(categories),
        )
        
        published = [a for a in articles if a.status == ArticleStatus.PUBLISHED]
        drafts = [a for a in articles if a.status == ArticleStatus.DRAFT]
        
        stats.published = len(published)
        stats.draft = len(drafts)
        stats.total_views = sum(a.view_count for a in articles)
        
        # Most viewed
        by_views = sorted(articles, key=lambda a: a.view_count, reverse=True)
        stats.most_viewed = [a.id for a in by_views[:5]]
        
        # Recently updated
        by_updated = sorted(articles, key=lambda a: a.updated_at, reverse=True)
        stats.recently_updated = [a.id for a in by_updated[:5]]
        
        return stats
    
    async def _save_version(self, article: Article, comment: str) -> None:
        """Save article version."""
        version = ArticleVersion(
            article_id=article.id,
            version=article.version,
            title=article.title,
            content=article.content,
            author=article.author,
            comment=comment,
        )
        await self._version_store.save(version)
    
    def add_listener(self, listener: Callable) -> None:
        """Add event listener."""
        self._listeners.append(listener)
    
    async def _notify(self, event: str, article: Article) -> None:
        """Notify listeners."""
        for listener in self._listeners:
            try:
                if asyncio.iscoroutinefunction(listener):
                    await listener(event, article)
                else:
                    listener(event, article)
            except Exception as e:
                logger.error(f"Listener error: {e}")


# Factory functions
def create_knowledge_manager() -> KnowledgeManager:
    """Create knowledge manager."""
    return KnowledgeManager()


def create_article(title: str, content: str, **kwargs) -> Article:
    """Create article."""
    return Article(title=title, content=content, **kwargs)


__all__ = [
    # Exceptions
    "KnowledgeError",
    # Enums
    "ArticleStatus",
    "ContentType",
    "ArticleType",
    # Data classes
    "Article",
    "Category",
    "SearchResult",
    "ArticleVersion",
    "KnowledgeStats",
    # Stores
    "ArticleStore",
    "InMemoryArticleStore",
    "CategoryStore",
    "InMemoryCategoryStore",
    "VersionStore",
    "InMemoryVersionStore",
    # Utilities
    "TextUtils",
    # Manager
    "KnowledgeManager",
    # Factory functions
    "create_knowledge_manager",
    "create_article",
]



"""
Enterprise Export Module.

Provides data export capabilities to various formats
including CSV, JSON, Excel, XML, and more.

Example:
    # Create exporter
    exporter = create_exporter("csv")
    
    # Export data
    await exporter.export(data, "output.csv")
    
    # With streaming for large datasets
    async with exporter.stream("large_output.csv") as writer:
        async for record in records:
            await writer.write(record)
    
    # Report generation
    report = create_report_builder()
    report.title("Sales Report")
    report.add_section("Summary", summary_data)
    report.add_table("Details", detail_rows)
    await report.export_pdf("report.pdf")
"""

from __future__ import annotations

import asyncio
import csv
import io
import json
import logging
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    BinaryIO,
    Callable,
    Dict,
    Generic,
    IO,
    Iterator,
    List,
    Optional,
    TextIO,
    Tuple,
    TypeVar,
    Union,
)

logger = logging.getLogger(__name__)

T = TypeVar('T')


class ExportError(Exception):
    """Export error."""
    pass


class FormatError(ExportError):
    """Format conversion error."""
    pass


class ExportFormat(str, Enum):
    """Export formats."""
    CSV = "csv"
    JSON = "json"
    JSONL = "jsonl"  # JSON Lines
    XML = "xml"
    HTML = "html"
    MARKDOWN = "markdown"
    YAML = "yaml"
    TSV = "tsv"
    EXCEL = "excel"
    PARQUET = "parquet"


class Compression(str, Enum):
    """Compression types."""
    NONE = "none"
    GZIP = "gzip"
    ZIP = "zip"
    BZIP2 = "bzip2"


@dataclass
class ExportConfig:
    """Export configuration."""
    format: ExportFormat = ExportFormat.JSON
    compression: Compression = Compression.NONE
    encoding: str = "utf-8"
    pretty: bool = True
    include_headers: bool = True
    date_format: str = "%Y-%m-%d %H:%M:%S"
    null_value: str = ""


@dataclass
class ExportResult:
    """Export result."""
    path: str
    format: ExportFormat
    size_bytes: int
    record_count: int
    duration_ms: float
    compressed: bool = False


@dataclass
class ColumnConfig:
    """Column configuration for tabular export."""
    name: str
    source: Optional[str] = None  # Source field name
    width: Optional[int] = None
    format: Optional[str] = None
    transform: Optional[Callable] = None
    header: Optional[str] = None  # Display header


class Exporter(ABC):
    """Abstract exporter."""
    
    @abstractmethod
    async def export(
        self,
        data: Any,
        destination: Union[str, IO],
    ) -> ExportResult:
        """Export data to destination."""
        pass
    
    @abstractmethod
    def export_sync(
        self,
        data: Any,
        destination: Union[str, IO],
    ) -> ExportResult:
        """Synchronous export."""
        pass


class CSVExporter(Exporter):
    """CSV exporter."""
    
    def __init__(
        self,
        config: Optional[ExportConfig] = None,
        columns: Optional[List[ColumnConfig]] = None,
    ):
        self._config = config or ExportConfig(format=ExportFormat.CSV)
        self._columns = columns
    
    async def export(
        self,
        data: List[Dict[str, Any]],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Export data to CSV."""
        return await asyncio.get_event_loop().run_in_executor(
            None, self.export_sync, data, destination
        )
    
    def export_sync(
        self,
        data: List[Dict[str, Any]],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Synchronous CSV export."""
        import time
        start = time.time()
        
        if not data:
            raise ExportError("No data to export")
        
        # Determine columns
        if self._columns:
            fieldnames = [c.header or c.name for c in self._columns]
            source_fields = [c.source or c.name for c in self._columns]
        else:
            source_fields = list(data[0].keys())
            fieldnames = source_fields
        
        # Handle file or IO
        should_close = False
        if isinstance(destination, str):
            f = open(destination, 'w', newline='', encoding=self._config.encoding)
            should_close = True
        else:
            f = destination
        
        try:
            writer = csv.writer(f)
            
            # Write header
            if self._config.include_headers:
                writer.writerow(fieldnames)
            
            # Write data
            for row in data:
                values = []
                for i, field_name in enumerate(source_fields):
                    value = row.get(field_name, self._config.null_value)
                    
                    # Apply transform
                    if self._columns and self._columns[i].transform:
                        value = self._columns[i].transform(value)
                    
                    # Format value
                    value = self._format_value(value)
                    values.append(value)
                
                writer.writerow(values)
            
            # Get file size
            if isinstance(destination, str):
                size = os.path.getsize(destination)
            else:
                size = f.tell() if hasattr(f, 'tell') else 0
            
            return ExportResult(
                path=destination if isinstance(destination, str) else "<stream>",
                format=ExportFormat.CSV,
                size_bytes=size,
                record_count=len(data),
                duration_ms=(time.time() - start) * 1000,
            )
        
        finally:
            if should_close:
                f.close()
    
    def _format_value(self, value: Any) -> str:
        """Format value for CSV."""
        if value is None:
            return self._config.null_value
        elif isinstance(value, datetime):
            return value.strftime(self._config.date_format)
        elif isinstance(value, (list, dict)):
            return json.dumps(value)
        return str(value)


class JSONExporter(Exporter):
    """JSON exporter."""
    
    def __init__(
        self,
        config: Optional[ExportConfig] = None,
    ):
        self._config = config or ExportConfig(format=ExportFormat.JSON)
    
    async def export(
        self,
        data: Any,
        destination: Union[str, IO],
    ) -> ExportResult:
        """Export data to JSON."""
        return await asyncio.get_event_loop().run_in_executor(
            None, self.export_sync, data, destination
        )
    
    def export_sync(
        self,
        data: Any,
        destination: Union[str, IO],
    ) -> ExportResult:
        """Synchronous JSON export."""
        import time
        start = time.time()
        
        indent = 2 if self._config.pretty else None
        
        json_str = json.dumps(
            data,
            indent=indent,
            default=self._json_serializer,
            ensure_ascii=False,
        )
        
        if isinstance(destination, str):
            with open(destination, 'w', encoding=self._config.encoding) as f:
                f.write(json_str)
            size = os.path.getsize(destination)
        else:
            destination.write(json_str)
            size = len(json_str.encode(self._config.encoding))
        
        record_count = len(data) if isinstance(data, list) else 1
        
        return ExportResult(
            path=destination if isinstance(destination, str) else "<stream>",
            format=ExportFormat.JSON,
            size_bytes=size,
            record_count=record_count,
            duration_ms=(time.time() - start) * 1000,
        )
    
    def _json_serializer(self, obj: Any) -> Any:
        """Custom JSON serializer."""
        if isinstance(obj, datetime):
            return obj.strftime(self._config.date_format)
        elif hasattr(obj, '__dict__'):
            return obj.__dict__
        raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


class JSONLExporter(Exporter):
    """JSON Lines exporter (one JSON object per line)."""
    
    def __init__(
        self,
        config: Optional[ExportConfig] = None,
    ):
        self._config = config or ExportConfig(format=ExportFormat.JSONL)
    
    async def export(
        self,
        data: List[Any],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Export data to JSON Lines."""
        return await asyncio.get_event_loop().run_in_executor(
            None, self.export_sync, data, destination
        )
    
    def export_sync(
        self,
        data: List[Any],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Synchronous JSONL export."""
        import time
        start = time.time()
        
        should_close = False
        if isinstance(destination, str):
            f = open(destination, 'w', encoding=self._config.encoding)
            should_close = True
        else:
            f = destination
        
        try:
            for item in data:
                line = json.dumps(item, default=str, ensure_ascii=False)
                f.write(line + '\n')
            
            if isinstance(destination, str):
                size = os.path.getsize(destination)
            else:
                size = f.tell() if hasattr(f, 'tell') else 0
            
            return ExportResult(
                path=destination if isinstance(destination, str) else "<stream>",
                format=ExportFormat.JSONL,
                size_bytes=size,
                record_count=len(data),
                duration_ms=(time.time() - start) * 1000,
            )
        finally:
            if should_close:
                f.close()


class XMLExporter(Exporter):
    """XML exporter."""
    
    def __init__(
        self,
        config: Optional[ExportConfig] = None,
        root_element: str = "data",
        row_element: str = "item",
    ):
        self._config = config or ExportConfig(format=ExportFormat.XML)
        self._root = root_element
        self._row = row_element
    
    async def export(
        self,
        data: Any,
        destination: Union[str, IO],
    ) -> ExportResult:
        """Export data to XML."""
        return await asyncio.get_event_loop().run_in_executor(
            None, self.export_sync, data, destination
        )
    
    def export_sync(
        self,
        data: Any,
        destination: Union[str, IO],
    ) -> ExportResult:
        """Synchronous XML export."""
        import time
        start = time.time()
        
        xml_str = self._to_xml(data)
        
        if isinstance(destination, str):
            with open(destination, 'w', encoding=self._config.encoding) as f:
                f.write(xml_str)
            size = os.path.getsize(destination)
        else:
            destination.write(xml_str)
            size = len(xml_str.encode(self._config.encoding))
        
        record_count = len(data) if isinstance(data, list) else 1
        
        return ExportResult(
            path=destination if isinstance(destination, str) else "<stream>",
            format=ExportFormat.XML,
            size_bytes=size,
            record_count=record_count,
            duration_ms=(time.time() - start) * 1000,
        )
    
    def _to_xml(self, data: Any) -> str:
        """Convert data to XML string."""
        lines = ['<?xml version="1.0" encoding="UTF-8"?>']
        lines.append(f'<{self._root}>')
        
        if isinstance(data, list):
            for item in data:
                lines.extend(self._element_to_xml(self._row, item, indent=2))
        else:
            lines.extend(self._element_to_xml(self._row, data, indent=2))
        
        lines.append(f'</{self._root}>')
        
        return '\n'.join(lines)
    
    def _element_to_xml(
        self,
        name: str,
        value: Any,
        indent: int = 0,
    ) -> List[str]:
        """Convert element to XML lines."""
        prefix = ' ' * indent
        
        if value is None:
            return [f'{prefix}<{name}/>']
        
        elif isinstance(value, dict):
            lines = [f'{prefix}<{name}>']
            for k, v in value.items():
                lines.extend(self._element_to_xml(k, v, indent + 2))
            lines.append(f'{prefix}</{name}>')
            return lines
        
        elif isinstance(value, list):
            lines = []
            for item in value:
                lines.extend(self._element_to_xml(name, item, indent))
            return lines
        
        else:
            escaped = self._escape_xml(str(value))
            return [f'{prefix}<{name}>{escaped}</{name}>']
    
    def _escape_xml(self, s: str) -> str:
        """Escape XML special characters."""
        return (s
            .replace('&', '&amp;')
            .replace('<', '&lt;')
            .replace('>', '&gt;')
            .replace('"', '&quot;')
            .replace("'", '&apos;'))


class MarkdownExporter(Exporter):
    """Markdown table exporter."""
    
    def __init__(
        self,
        config: Optional[ExportConfig] = None,
        columns: Optional[List[ColumnConfig]] = None,
    ):
        self._config = config or ExportConfig(format=ExportFormat.MARKDOWN)
        self._columns = columns
    
    async def export(
        self,
        data: List[Dict[str, Any]],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Export data to Markdown table."""
        return await asyncio.get_event_loop().run_in_executor(
            None, self.export_sync, data, destination
        )
    
    def export_sync(
        self,
        data: List[Dict[str, Any]],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Synchronous Markdown export."""
        import time
        start = time.time()
        
        if not data:
            raise ExportError("No data to export")
        
        # Determine columns
        if self._columns:
            headers = [c.header or c.name for c in self._columns]
            fields = [c.source or c.name for c in self._columns]
        else:
            fields = list(data[0].keys())
            headers = fields
        
        # Build markdown table
        lines = []
        
        # Header row
        lines.append('| ' + ' | '.join(headers) + ' |')
        lines.append('| ' + ' | '.join(['---'] * len(headers)) + ' |')
        
        # Data rows
        for row in data:
            values = []
            for field_name in fields:
                value = row.get(field_name, '')
                values.append(str(value).replace('|', '\\|'))
            lines.append('| ' + ' | '.join(values) + ' |')
        
        content = '\n'.join(lines)
        
        if isinstance(destination, str):
            with open(destination, 'w', encoding=self._config.encoding) as f:
                f.write(content)
            size = os.path.getsize(destination)
        else:
            destination.write(content)
            size = len(content.encode(self._config.encoding))
        
        return ExportResult(
            path=destination if isinstance(destination, str) else "<stream>",
            format=ExportFormat.MARKDOWN,
            size_bytes=size,
            record_count=len(data),
            duration_ms=(time.time() - start) * 1000,
        )


class HTMLExporter(Exporter):
    """HTML table exporter."""
    
    def __init__(
        self,
        config: Optional[ExportConfig] = None,
        title: str = "Export",
        css: Optional[str] = None,
    ):
        self._config = config or ExportConfig(format=ExportFormat.HTML)
        self._title = title
        self._css = css or self._default_css()
    
    def _default_css(self) -> str:
        return """
        body { font-family: Arial, sans-serif; margin: 20px; }
        table { border-collapse: collapse; width: 100%; }
        th, td { border: 1px solid #ddd; padding: 8px; text-align: left; }
        th { background-color: #4CAF50; color: white; }
        tr:nth-child(even) { background-color: #f2f2f2; }
        tr:hover { background-color: #ddd; }
        """
    
    async def export(
        self,
        data: List[Dict[str, Any]],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Export data to HTML."""
        return await asyncio.get_event_loop().run_in_executor(
            None, self.export_sync, data, destination
        )
    
    def export_sync(
        self,
        data: List[Dict[str, Any]],
        destination: Union[str, IO],
    ) -> ExportResult:
        """Synchronous HTML export."""
        import time
        start = time.time()
        
        if not data:
            raise ExportError("No data to export")
        
        headers = list(data[0].keys())
        
        html_parts = [
            '<!DOCTYPE html>',
            '<html><head>',
            f'<title>{self._title}</title>',
            f'<style>{self._css}</style>',
            '</head><body>',
            f'<h1>{self._title}</h1>',
            '<table>',
            '<thead><tr>',
        ]
        
        # Headers
        for h in headers:
            html_parts.append(f'<th>{h}</th>')
        html_parts.append('</tr></thead>')
        
        # Body
        html_parts.append('<tbody>')
        for row in data:
            html_parts.append('<tr>')
            for h in headers:
                value = row.get(h, '')
                html_parts.append(f'<td>{self._escape_html(str(value))}</td>')
            html_parts.append('</tr>')
        html_parts.append('</tbody>')
        
        html_parts.extend(['</table>', '</body></html>'])
        
        content = '\n'.join(html_parts)
        
        if isinstance(destination, str):
            with open(destination, 'w', encoding=self._config.encoding) as f:
                f.write(content)
            size = os.path.getsize(destination)
        else:
            destination.write(content)
            size = len(content.encode(self._config.encoding))
        
        return ExportResult(
            path=destination if isinstance(destination, str) else "<stream>",
            format=ExportFormat.HTML,
            size_bytes=size,
            record_count=len(data),
            duration_ms=(time.time() - start) * 1000,
        )
    
    def _escape_html(self, s: str) -> str:
        """Escape HTML special characters."""
        return (s
            .replace('&', '&amp;')
            .replace('<', '&lt;')
            .replace('>', '&gt;')
            .replace('"', '&quot;'))


class StreamingExporter:
    """
    Streaming exporter for large datasets.
    """
    
    def __init__(
        self,
        exporter: Exporter,
        chunk_size: int = 1000,
    ):
        self._exporter = exporter
        self._chunk_size = chunk_size
    
    async def export_iterator(
        self,
        iterator: AsyncIterator[Dict[str, Any]],
        destination: str,
    ) -> ExportResult:
        """Export from async iterator."""
        import time
        start = time.time()
        
        chunk = []
        total_records = 0
        
        # First chunk determines headers
        first_chunk = True
        
        async for item in iterator:
            chunk.append(item)
            
            if len(chunk) >= self._chunk_size:
                await self._write_chunk(chunk, destination, first_chunk)
                total_records += len(chunk)
                chunk = []
                first_chunk = False
        
        # Write remaining
        if chunk:
            await self._write_chunk(chunk, destination, first_chunk)
            total_records += len(chunk)
        
        size = os.path.getsize(destination) if os.path.exists(destination) else 0
        
        return ExportResult(
            path=destination,
            format=ExportFormat.CSV,
            size_bytes=size,
            record_count=total_records,
            duration_ms=(time.time() - start) * 1000,
        )
    
    async def _write_chunk(
        self,
        chunk: List[Dict[str, Any]],
        destination: str,
        include_headers: bool,
    ) -> None:
        """Write a chunk to file."""
        mode = 'w' if include_headers else 'a'
        
        with open(destination, mode, newline='') as f:
            if chunk:
                writer = csv.DictWriter(f, fieldnames=chunk[0].keys())
                if include_headers:
                    writer.writeheader()
                writer.writerows(chunk)


class ReportBuilder:
    """
    Build reports with multiple sections.
    """
    
    def __init__(self, title: str = "Report"):
        self._title = title
        self._sections: List[Dict[str, Any]] = []
        self._metadata: Dict[str, Any] = {
            "created_at": datetime.now(),
            "author": "",
        }
    
    def title(self, title: str) -> "ReportBuilder":
        """Set report title."""
        self._title = title
        return self
    
    def author(self, author: str) -> "ReportBuilder":
        """Set author."""
        self._metadata["author"] = author
        return self
    
    def add_section(
        self,
        name: str,
        content: Any,
        section_type: str = "text",
    ) -> "ReportBuilder":
        """Add a section to the report."""
        self._sections.append({
            "name": name,
            "content": content,
            "type": section_type,
        })
        return self
    
    def add_table(
        self,
        name: str,
        data: List[Dict[str, Any]],
        columns: Optional[List[str]] = None,
    ) -> "ReportBuilder":
        """Add a table section."""
        self._sections.append({
            "name": name,
            "content": data,
            "type": "table",
            "columns": columns,
        })
        return self
    
    def add_chart(
        self,
        name: str,
        chart_type: str,
        data: Any,
    ) -> "ReportBuilder":
        """Add a chart section (placeholder for visualization)."""
        self._sections.append({
            "name": name,
            "content": data,
            "type": "chart",
            "chart_type": chart_type,
        })
        return self
    
    async def export_html(self, path: str) -> ExportResult:
        """Export report as HTML."""
        html = self._build_html()
        
        with open(path, 'w') as f:
            f.write(html)
        
        return ExportResult(
            path=path,
            format=ExportFormat.HTML,
            size_bytes=os.path.getsize(path),
            record_count=len(self._sections),
            duration_ms=0,
        )
    
    async def export_markdown(self, path: str) -> ExportResult:
        """Export report as Markdown."""
        md = self._build_markdown()
        
        with open(path, 'w') as f:
            f.write(md)
        
        return ExportResult(
            path=path,
            format=ExportFormat.MARKDOWN,
            size_bytes=os.path.getsize(path),
            record_count=len(self._sections),
            duration_ms=0,
        )
    
    def _build_html(self) -> str:
        """Build HTML report."""
        parts = [
            '<!DOCTYPE html>',
            '<html><head>',
            f'<title>{self._title}</title>',
            '<style>',
            'body { font-family: Arial; max-width: 1200px; margin: 0 auto; padding: 20px; }',
            'table { border-collapse: collapse; width: 100%; margin: 20px 0; }',
            'th, td { border: 1px solid #ddd; padding: 8px; }',
            'th { background: #f5f5f5; }',
            '</style>',
            '</head><body>',
            f'<h1>{self._title}</h1>',
            f'<p>Generated: {self._metadata["created_at"]}</p>',
        ]
        
        for section in self._sections:
            parts.append(f'<h2>{section["name"]}</h2>')
            
            if section["type"] == "table":
                parts.append(self._table_to_html(section["content"]))
            elif section["type"] == "text":
                parts.append(f'<p>{section["content"]}</p>')
        
        parts.extend(['</body></html>'])
        return '\n'.join(parts)
    
    def _build_markdown(self) -> str:
        """Build Markdown report."""
        parts = [
            f'# {self._title}',
            '',
            f'*Generated: {self._metadata["created_at"]}*',
            '',
        ]
        
        for section in self._sections:
            parts.append(f'## {section["name"]}')
            parts.append('')
            
            if section["type"] == "table":
                parts.append(self._table_to_markdown(section["content"]))
            elif section["type"] == "text":
                parts.append(str(section["content"]))
            
            parts.append('')
        
        return '\n'.join(parts)
    
    def _table_to_html(self, data: List[Dict]) -> str:
        """Convert table data to HTML."""
        if not data:
            return '<p>No data</p>'
        
        headers = list(data[0].keys())
        rows = ['<table><thead><tr>']
        rows.extend(f'<th>{h}</th>' for h in headers)
        rows.append('</tr></thead><tbody>')
        
        for row in data:
            rows.append('<tr>')
            for h in headers:
                rows.append(f'<td>{row.get(h, "")}</td>')
            rows.append('</tr>')
        
        rows.append('</tbody></table>')
        return ''.join(rows)
    
    def _table_to_markdown(self, data: List[Dict]) -> str:
        """Convert table data to Markdown."""
        if not data:
            return 'No data'
        
        headers = list(data[0].keys())
        lines = ['| ' + ' | '.join(headers) + ' |']
        lines.append('| ' + ' | '.join(['---'] * len(headers)) + ' |')
        
        for row in data:
            values = [str(row.get(h, '')) for h in headers]
            lines.append('| ' + ' | '.join(values) + ' |')
        
        return '\n'.join(lines)


# Factory functions
def create_exporter(
    format: Union[str, ExportFormat],
    **kwargs: Any,
) -> Exporter:
    """Create an exporter for the specified format."""
    if isinstance(format, str):
        format = ExportFormat(format.lower())
    
    exporters = {
        ExportFormat.CSV: CSVExporter,
        ExportFormat.JSON: JSONExporter,
        ExportFormat.JSONL: JSONLExporter,
        ExportFormat.XML: XMLExporter,
        ExportFormat.MARKDOWN: MarkdownExporter,
        ExportFormat.HTML: HTMLExporter,
    }
    
    if format not in exporters:
        raise ExportError(f"Unsupported format: {format}")
    
    return exporters[format](**kwargs)


def create_report_builder(title: str = "Report") -> ReportBuilder:
    """Create a report builder."""
    return ReportBuilder(title)


def create_streaming_exporter(
    format: Union[str, ExportFormat] = "csv",
    chunk_size: int = 1000,
) -> StreamingExporter:
    """Create a streaming exporter."""
    base = create_exporter(format)
    return StreamingExporter(base, chunk_size)


# Decorators
def export_result(
    format: ExportFormat = ExportFormat.JSON,
    path_fn: Optional[Callable] = None,
) -> Callable:
    """
    Decorator to export function result.
    
    Example:
        @export_result(ExportFormat.CSV, path_fn=lambda: f"output_{datetime.now()}.csv")
        async def generate_report():
            return [{"a": 1}, {"a": 2}]
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> ExportResult:
            result = await func(*args, **kwargs)
            
            path = path_fn() if path_fn else f"export_{datetime.now().timestamp()}.{format.value}"
            
            exporter = create_exporter(format)
            return await exporter.export(result, path)
        
        return wrapper
    
    return decorator


__all__ = [
    # Exceptions
    "ExportError",
    "FormatError",
    # Enums
    "ExportFormat",
    "Compression",
    # Data classes
    "ExportConfig",
    "ExportResult",
    "ColumnConfig",
    # Core classes
    "Exporter",
    "CSVExporter",
    "JSONExporter",
    "JSONLExporter",
    "XMLExporter",
    "MarkdownExporter",
    "HTMLExporter",
    "StreamingExporter",
    "ReportBuilder",
    # Factory functions
    "create_exporter",
    "create_report_builder",
    "create_streaming_exporter",
    # Decorators
    "export_result",
]
